from streamlink.plugins.tv4play import TV4Play
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV4Play(PluginCanHandleUrl):
    __plugin__ = TV4Play

    should_match = [
        'https://www.tv4play.se/program/robinson/del-26-sasong-2021/13299862',
        'https://www.tv4play.se/program/sverige-mot-norge/del-1-sasong-1/12490380',
        'https://www.tv4play.se/program/nyheterna/live/10378590',
        'https://www.fotbollskanalen.se/video/10395484/ghoddos-fullbordar-vandningen---ger-ofk-ledningen/',
    ]



"""
$description Live TV channels and video on-demand service from UseeTV, owned by Telkom Indonesia.
$url useetv.com
$type live, vod
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"https?://(?:www\.)?useetv\.com/"))
class UseeTV(Plugin):
    def _get_streams(self):
        root = self.session.http.get(self.url, schema=validate.Schema(validate.parse_html()))

        for needle, errormsg in (
            (
                "\"This service is not available in your Country\"",
                "The content is not available in your region",
            ),
            (
                "\"Silahkan login Menggunakan akun MyIndihome dan berlangganan minipack\"",
                "The content is not available without a subscription",
            ),
        ):
            if validate.Schema(validate.xml_xpath(".//script[contains(text(),$needle)]", needle=needle)).validate(root):
                log.error(errormsg)
                return

        url = validate.Schema(
            validate.any(
                validate.all(
                    validate.xml_xpath_string("""
                        .//script[contains(text(), 'laylist.m3u8') or contains(text(), 'manifest.mpd')][1]/text()
                    """),
                    str,
                    re.compile(r"""(?P<q>['"])(?P<url>https://.*?/(?:[Pp]laylist\.m3u8|manifest\.mpd).+?)(?P=q)"""),
                    validate.none_or_all(validate.get("url"), validate.url()),
                ),
                validate.all(
                    validate.xml_xpath_string(".//video[@id='video-player']/source/@src"),
                    validate.any(None, validate.url()),
                ),
            )
        ).validate(root)

        if url and ".m3u8" in url:
            return HLSStream.parse_variant_playlist(self.session, url)
        elif url and ".mpd" in url:
            return DASHStream.parse_manifest(self.session, url)


__plugin__ = UseeTV



"""
$description Russian live TV channel owned by Gazprom Media.
$url ntv.ru
$type live
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r'https?://www\.ntv\.ru/air/'
))
class NTV(Plugin):
    def _get_streams(self):
        body = self.session.http.get(self.url).text
        mrl = None
        match = re.search(r'var camHlsURL = \'(.*)\'', body)
        if match:
            mrl = f"http:{match.group(1)}"
        else:
            match = re.search(r'var hlsURL = \'(.*)\'', body)
            if match:
                mrl = match.group(1)
        if mrl:
            return HLSStream.parse_variant_playlist(self.session, mrl)


__plugin__ = NTV



from streamlink.plugins.ruv import Ruv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRuv(PluginCanHandleUrl):
    __plugin__ = Ruv

    should_match = [
        "http://ruv.is/ruv",
        "http://ruv.is/ruv",
        "http://ruv.is/ruv/",
        "https://ruv.is/ruv/",
        "http://www.ruv.is/ruv",
        "http://www.ruv.is/ruv/",
        "http://ruv.is/ruv2",
        "http://ruv.is/ras1",
        "http://ruv.is/ras2",
        "http://ruv.is/rondo",
        "http://www.ruv.is/spila/ruv/ol-2018-ishokki-karla/20180217",
        "http://www.ruv.is/spila/ruv/frettir/20180217"
    ]

    should_not_match = [
        "http://ruv.is/ruvnew",
    ]



from streamlink.plugins.nimotv import NimoTV
from tests.plugins import PluginCanHandleUrl


class TestPluginNimoTV(PluginCanHandleUrl):
    __plugin__ = NimoTV

    should_match = [
        'http://www.nimo.tv/live/737614',
        'https://www.nimo.tv/live/737614',
        'http://www.nimo.tv/sanz',
        'https://www.nimo.tv/sanz',
        'https://m.nimo.tv/user',
    ]



import abc
import ast
import pkgutil
import re
import tokenize
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from sphinx.errors import ExtensionError
from sphinx.util.nodes import nested_parse_with_titles

from streamlink import plugins as streamlink_plugins


class IDatalistItem(abc.ABC):
    @abc.abstractmethod
    def generate(self) -> Iterator[str]:
        raise NotImplementedError


class IMetadataItem(IDatalistItem):
    @abc.abstractmethod
    def set(self, value: Any) -> None:
        raise NotImplementedError


class MetadataItem(IMetadataItem):
    def __init__(self, title: str):
        self.title = title
        self.value: Optional[str] = None

    def set(self, value: str) -> None:
        self.value = value

    def generate(self) -> Iterator[str]:
        if self.value is None:
            return
        yield f":{self.title}: {self.value}"


class MetadataList(IMetadataItem):
    def __init__(self, title: str):
        self.title = title
        self.value: List[str] = []

    def set(self, value: str) -> None:
        self.value.append(value)

    def generate(self) -> Iterator[str]:
        if not self.value:
            return
        indent = " " * len(f":{self.title}:")
        yield f":{self.title}: - {self.value[0]}"
        for val in self.value[1:]:
            yield f"{indent} - {val}"


class PluginOnGithub(IDatalistItem):
    url_source = "https://github.com/streamlink/streamlink/blob/master/src/streamlink/plugins/{name}.py"
    url_issues = "https://github.com/streamlink/streamlink/issues?q=is%3Aissue+sort%3Aupdated-desc+plugins.{name}"

    def __init__(self, pluginname: str):
        self.pluginname = pluginname

    def generate(self) -> Iterator[str]:
        source = f"`Source <{self.url_source.format(name=self.pluginname)}>`__"
        issues = f"`Issues <{self.url_issues.format(name=self.pluginname)}>`__"
        yield f":GitHub: {source}, {issues}"


class PluginArguments(ast.NodeVisitor, IDatalistItem):
    def __init__(self, pluginname, pluginast):
        super().__init__()
        self.pluginname = pluginname
        self.arguments = []
        self.visit(pluginast)

    def generate(self) -> Iterator[str]:
        if not self.arguments:
            return
        indent = " " * len(":Arguments: ")
        yield f":Arguments: - :option:`--{self.arguments[0]}`"
        for arg in self.arguments[1:]:
            yield f"{indent} - :option:`--{arg}`"

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        for base in node.bases:
            if not isinstance(base, ast.Name):
                continue
            if base.id == "Plugin":
                break
        else:
            return

        for decorator in node.decorator_list:
            if (
                not isinstance(decorator, ast.Call)
                or not isinstance(decorator.func, ast.Name)
                or decorator.func.id != "pluginargument"
                or (len(decorator.args) == 0 and len(decorator.keywords) == 0)
            ):
                continue

            if any(kw.value.value for kw in decorator.keywords if kw.arg == "is_global" and type(kw.value) is ast.Constant):
                continue

            if any(
                True
                for kw in decorator.keywords
                if (
                    kw.arg == "help"
                    and type(kw.value) is ast.Attribute
                    and kw.value.attr == "SUPPRESS"
                    and type(kw.value.value) is ast.Name
                    and kw.value.value.id == "argparse"
                )
            ):
                continue

            custom_name = next(
                (kw.value.value for kw in decorator.keywords if kw.arg == "argument_name" and type(kw.value) is ast.Constant),
                None,
            )
            if custom_name:
                self.arguments.append(custom_name)
                continue

            name = next(
                (kw.value.value for kw in decorator.keywords if kw.arg == "name" and type(kw.value) is ast.Constant),
                None,
            ) or (
                decorator.args
                and type(decorator.args[0]) is ast.Constant
                and decorator.args[0].value
            )
            if name:
                self.arguments.append(f"{self.pluginname}-{name}")


class PluginMetadata:
    def __init__(self, name: str, pluginast):
        self.name: str = name
        self.items: Dict[str, IMetadataItem] = dict(
            description=MetadataItem("Description"),
            url=MetadataList("URL(s)"),
            type=MetadataItem("Type"),
            region=MetadataItem("Region"),
            account=MetadataItem("Account"),
            notes=MetadataList("Notes"),
        )
        self.additional: List[IDatalistItem] = [
            PluginArguments(name, pluginast),
            PluginOnGithub(name),
        ]

    def set(self, key: str, value: str) -> None:
        if key not in self.items:
            raise KeyError(f"Invalid plugin metadata key '{key}' in plugin '{self.name}'")
        self.items[key].set(value)

    def generate(self) -> Iterator[str]:
        yield self.name
        yield "-" * len(self.name)
        yield ""
        for metadata in self.items.values():
            yield from metadata.generate()
        for item in self.additional:
            yield from item.generate()
        yield ""
        yield ""


class PluginFinder:
    _re_metadata_item = re.compile(r"\n\$(\w+) (.+)(?=\n\$|$)", re.MULTILINE)

    protocol_plugins = [
        "http",
        "hls",
        "dash",
    ]

    def __init__(self):
        plugins_path = Path(streamlink_plugins.__path__[0])
        self.plugins = [
            (pname, plugins_path / f"{pname}.py")
            for finder, pname, ispkg in pkgutil.iter_modules([str(plugins_path)])
            if not pname.startswith("common_") and pname not in self.protocol_plugins
        ]

    def get_plugins(self):
        for pluginname, pluginfile in self.plugins:
            pluginmetadata = self._parse_plugin(pluginname, pluginfile)
            if pluginmetadata:
                yield pluginmetadata

    def _parse_plugin(self, pluginname: str, pluginfile: Path) -> Optional[PluginMetadata]:
        with pluginfile.open() as handle:
            # read until the first token has been parsed
            for tokeninfo in tokenize.generate_tokens(handle.readline):
                # the very first token needs to be a string / block comment with the metadata
                if tokeninfo.type != tokenize.STRING or not self._re_metadata_item.search(tokeninfo.string):
                    return None
                metadata = tokeninfo.string.strip()
                break

            try:
                # continue reading the plugin file with the same handle
                pluginsource = handle.read()
                # build AST from plugin source for finding the used plugin arguments
                pluginast = ast.parse(pluginsource, str(pluginfile))

                pluginmetadata = PluginMetadata(pluginname, pluginast)
                for item in self._re_metadata_item.finditer(metadata):
                    key, value = item.groups()
                    pluginmetadata.set(key, value)

                return pluginmetadata

            except Exception as err:
                raise ExtensionError(f"Error while parsing plugin file {pluginfile.name}", err)


class PluginsDirective(Directive):
    def run(self):
        pluginfinder = PluginFinder()

        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        for pluginmetadata in pluginfinder.get_plugins():
            for line in pluginmetadata.generate():
                result.append(line, "plugins")

        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    app.add_directive("plugins", PluginsDirective)



from streamlink.plugins.stadium import Stadium
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlStadium(PluginCanHandleUrl):
    __plugin__ = Stadium

    should_match = [
        "http://www.watchstadium.com/live",
        "https://www.watchstadium.com/live",
        "https://watchstadium.com/live",
        "http://watchstadium.com/live",
        "https://watchstadium.com/sport/college-football/",
    ]



import os
import sys


is_darwin = sys.platform == "darwin"
is_win32 = os.name == "nt"

# win/nix compatible devnull
try:
    from subprocess import DEVNULL

    def devnull():
        return DEVNULL
except ImportError:
    def devnull():
        return open(os.path.devnull, 'w')


__all__ = [
    "is_darwin",
    "is_win32",
    "devnull",
]



from streamlink.plugins.mdstrm import MDStrm
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMDStrm(PluginCanHandleUrl):
    __plugin__ = MDStrm

    should_match = [
        "https://mdstrm.com/live-stream/57b4dbf5dbbfc8f16bb63ce1",
        "https://mdstrm.com/live-stream/5a7b1e63a8da282c34d65445",
        "https://mdstrm.com/live-stream/5ce7109c7398b977dc0744cd",
        "https://mdstrm.com/live-stream/60b578b060947317de7b57ac",
        "https://mdstrm.com/live-stream/61e1e088d04d7744686afc42",
        "https://saltillo.multimedios.com/video/monterrey-tv-en-vivo/v7567",
        "https://www.latina.pe/tvenvivo",
    ]



from datetime import datetime, timedelta, timezone
from json import JSONDecodeError
from pathlib import Path
from typing import Type, Union
from unittest.mock import Mock, patch

import freezegun
import pytest

from streamlink.cache import Cache


@pytest.fixture(autouse=True)
def cache_dir(tmp_path: Path):
    with patch("streamlink.cache.CACHE_DIR", tmp_path):
        yield tmp_path


@pytest.fixture
def cache(request: pytest.FixtureRequest, cache_dir: Path):
    param = getattr(request, "param", {})
    filename = param.get("filename", "cache.json")
    key_prefix = param.get("key_prefix", None)

    cache = Cache(filename, key_prefix=key_prefix)
    assert cache.filename == cache_dir / filename
    # noinspection PyProtectedMember
    assert not cache._cache

    return cache


class TestPathlibAndStr:
    @pytest.mark.parametrize("filename", [
        pytest.param("foo", id="str"),
        pytest.param(Path("foo"), id="Path"),
    ])
    def test_constructor(self, cache_dir: Path, filename: Union[str, Path]):
        cache = Cache(filename)
        assert cache.filename == cache_dir / Path(filename)


class TestGetterSetter:
    def test_get(self, cache: Cache):
        assert cache.get("missing-value") is None
        assert cache.get("missing-value", default="default") == "default"

    def test_set(self, cache: Cache):
        assert cache.get("value") is None
        cache.set("value", 1)
        assert cache.get("value") == 1
        assert cache._cache

    def test_get_all(self, cache: Cache):
        cache.set("test1", 1)
        cache.set("test2", 2)
        assert cache.get_all() == {"test1": 1, "test2": 2}

    def test_get_all_prune(self, cache: Cache):
        cache.set("test1", 1)
        cache.set("test2", 2, -1)
        assert cache.get_all() == {"test1": 1}


class TestPrefix:
    @pytest.mark.parametrize("cache", [{"key_prefix": "test"}], indirect=["cache"])
    def test_key_prefix(self, cache: Cache):
        cache.set("key", 1)
        assert cache.get("key") == 1
        assert "test:key" in cache._cache
        assert cache._cache["test:key"]["value"] == 1

    def test_get_all_prefix(self, cache: Cache):
        cache.set("test1", 1)
        cache.set("test2", 2)
        cache.key_prefix = "test"
        cache.set("test3", 3)
        cache.set("test4", 4)
        assert cache.get_all() == {"test3": 3, "test4": 4}


class TestExpiration:
    @pytest.mark.parametrize("expires,expected", [
        pytest.param(-20, None, id="past"),
        pytest.param(20, "value", id="future"),
    ])
    def test_expires(self, cache: Cache, expires: float, expected):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache.set("key", "value", expires=expires)
            assert cache.get("key") == expected

    @pytest.mark.parametrize("delta,expected", [
        pytest.param(timedelta(seconds=-20), None, id="past"),
        pytest.param(timedelta(seconds=20), "value", id="future"),
    ])
    def test_expires_at(self, cache: Cache, delta: timedelta, expected):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache.set("key", "value", expires_at=datetime.now(tz=timezone.utc) + delta)
            assert cache.get("key") == expected

    def test_expires_at_overflowerror(self, cache: Cache):
        expires_at = Mock(timestamp=Mock(side_effect=OverflowError))
        cache.set("key", "value", expires_at=expires_at)
        assert cache.get("key") is None

    def test_expiration(self, cache: Cache):
        with freezegun.freeze_time("2000-01-01T00:00:00Z") as frozen_time:
            cache.set("key", "value", expires=20)
            assert cache.get("key") == "value"
            frozen_time.tick(timedelta(seconds=20))
            assert cache.get("key") is None


class TestIO:
    @pytest.mark.parametrize("mockpath,side_effect", [
        ("pathlib.Path.open", OSError),
        ("json.load", JSONDecodeError),
    ])
    def test_load_fail(self, cache: Cache, mockpath: str, side_effect: Type[Exception]):
        with patch("pathlib.Path.exists", return_value=True):
            with patch(mockpath, side_effect=side_effect):
                cache._load()
        assert not cache._cache

    @pytest.mark.parametrize("side_effect", [
        RecursionError,
        TypeError,
        ValueError,
    ])
    def test_save_fail_jsondump(self, cache: Cache, side_effect: Type[Exception]):
        with pytest.raises(side_effect):
            with patch("json.dump", side_effect=side_effect):
                cache.set("key", "value")
        assert not cache.filename.exists()


class TestCreateDirectory:
    filepath = Path("dir1", "dir2", "cache.json")

    def test_success(self, cache_dir: Path):
        expected = cache_dir / self.filepath
        cache = Cache(self.filepath)
        assert not expected.exists()
        cache.set("key", "value")
        assert expected.exists()

    def test_failure(self, cache_dir: Path):
        with patch("pathlib.Path.mkdir", side_effect=OSError):
            expected = cache_dir / self.filepath
            cache = Cache(self.filepath)
            assert not expected.exists()
            cache.set("key", "value")
            assert not expected.exists()
            assert not list(cache_dir.iterdir())



"""
$description Global and country-specific websites for live radio simulcasts for over 40,000 stations.
$url radio.net
$url radio.at
$url radio.de
$url radio.dk
$url radio.es
$url radio.fr
$url radio.it
$url radio.pl
$url radio.pt
$url radio.se
$type live
"""

import logging
import re
from urllib.parse import urlparse, urlunparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(\w+)\.radio\.(net|at|de|dk|es|fr|it|pl|pt|se)"
))
class RadioNet(Plugin):
    def _get_streams(self):
        streams = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"\bstation\s*:\s*(\{.+}),?\s*"),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                {
                    "type": str,
                    "streams": [{
                        "url": validate.url(),
                        "contentFormat": str,
                    }],
                },
            ),
        ))
        if streams is None:
            return

        if streams['type'] != 'STATION':
            return

        stream_urls = set()
        for stream in streams['streams']:
            log.trace('{0!r}'.format(stream))
            url = stream['url']

            url_no_scheme = urlunparse(urlparse(url)._replace(scheme=''))
            if url_no_scheme in stream_urls:
                continue
            stream_urls.add(url_no_scheme)

            if stream['contentFormat'] in ('audio/mpeg', 'audio/aac'):
                yield 'live', HTTPStream(self.session, url, allow_redirects=True)
            elif stream['contentFormat'] == 'video/MP2T':
                streams = HLSStream.parse_variant_playlist(self.session, stream["url"])
                if not streams:
                    yield stream["quality"], HLSStream(self.session, stream["url"])
                else:
                    yield from streams.items()


__plugin__ = RadioNet



import os.path
import sys
import unittest

from streamlink.utils.module import load_module

# used in the import test to verify that this module was imported
__test_marker__ = "test_marker"


class TestUtilsModule(unittest.TestCase):
    def test_load_module_non_existent(self):
        self.assertRaises(ImportError, load_module, "non_existent_module", os.path.dirname(__file__))

    def test_load_module(self):
        self.assertEqual(
            sys.modules[__name__].__test_marker__,
            load_module(__name__.split(".")[-1], os.path.dirname(__file__)).__test_marker__
        )



from streamlink.plugins.zhanqi import Zhanqitv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlZhanqitv(PluginCanHandleUrl):
    __plugin__ = Zhanqitv

    should_match = [
        'https://www.zhanqi.tv/lpl',
    ]



"""
$description Live TV channels from DR, a Danish public, state-owned broadcaster.
$url dr.dk
$type live
$region Denmark
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?dr\.dk/drtv(/kanal/[\w-]+)"
))
class DRDK(Plugin):
    live_api_url = 'https://www.dr-massive.com/api/page'

    _live_data_schema = validate.Schema(
        {'item': {'customFields': {
            validate.optional('hlsURL'): validate.url(),
            validate.optional('hlsWithSubtitlesURL'): validate.url(),
        }}},
        validate.get('item'),
        validate.get('customFields'),
    )

    def _get_live(self, path):
        params = dict(
            ff='idp',
            path=path,
        )
        res = self.session.http.get(self.live_api_url, params=params)
        playlists = self.session.http.json(res, schema=self._live_data_schema)

        streams = {}
        for name, url in playlists.items():
            name_prefix = ''
            if name == 'hlsWithSubtitlesURL':
                name_prefix = 'subtitled_'

            streams.update(HLSStream.parse_variant_playlist(
                self.session,
                url,
                name_prefix=name_prefix,
            ))

        return streams

    def _get_streams(self):
        path = self.match.group(1)
        log.debug("Path={0}".format(path))

        return self._get_live(path)


__plugin__ = DRDK



"""
$description Russian live streaming platform for live video game broadcasts.
$url goodgame.ru
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream
from streamlink.utils.parse import parse_json

log = logging.getLogger(__name__)

HLS_URL_FORMAT = "https://hls.goodgame.ru/hls/{0}{1}.m3u8"
QUALITIES = {
    "1080p": "",
    "720p": "_720",
    "480p": "_480",
    "240p": "_240"
}

_apidata_re = re.compile(r'''(?P<quote>["']?)channel(?P=quote)\s*:\s*(?P<data>{.*?})\s*,''')
_ddos_re = re.compile(r'document.cookie="(__DDOS_[^;]+)')


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?goodgame\.ru/channel/(?P<user>[^/]+)"
))
class GoodGame(Plugin):
    def _check_stream(self, url):
        res = self.session.http.get(url, acceptable_status=(200, 404))
        if res.status_code == 200:
            return True

    def _get_streams(self):
        headers = {
            "Referer": self.url
        }
        res = self.session.http.get(self.url, headers=headers)

        match = _ddos_re.search(res.text)
        if match:
            log.debug("Anti-DDOS bypass...")
            headers["Cookie"] = match.group(1)
            res = self.session.http.get(self.url, headers=headers)

        match = _apidata_re.search(res.text)
        channel_info = match and parse_json(match.group("data"))
        if not channel_info:
            log.error("Could not find channel info")
            return

        log.debug("Found channel info: id={id} channelkey={channelkey} pid={streamkey} online={status}".format(**channel_info))
        if not channel_info['status']:
            log.debug("Channel appears to be offline")

        streams = {}
        for name, url_suffix in QUALITIES.items():
            url = HLS_URL_FORMAT.format(channel_info['streamkey'], url_suffix)
            if not self._check_stream(url):
                continue

            streams[name] = HLSStream(self.session, url)

        return streams


__plugin__ = GoodGame



"""
$description Live TV channels and video on-demand service from NRK, a Norwegian public, state-owned broadcaster.
$url tv.nrk.no
$url radio.nrk.no
$type live, vod
$region Norway
"""

import logging
import re
from urllib.parse import urljoin

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:tv|radio)\.nrk\.no/(program|direkte|serie|podkast)(?:/.+)?/([^/]+)"
))
class NRK(Plugin):
    _psapi_url = 'https://psapi.nrk.no'
    # Program type to manifest type mapping
    _program_type_map = {
        'direkte': 'channel',
        'serie': 'program',
        'program': 'program',
        'podkast': 'podcast',
    }

    _program_id_re = re.compile(r'<meta property="nrk:program-id" content="([^"]+)"')

    _playable_schema = validate.Schema(validate.any(
        {
            "playable": {
                "assets": [{
                    "url": validate.url(),
                    "format": str,
                }],
            },
            "statistics": {
                validate.optional("luna"): validate.any(None, {
                    "data": {
                        "title": str,
                        "category": validate.any(None, str),
                    },
                }),
            },
        },
        {
            "nonPlayable": {
                "reason": str,
            },
        },
    ))

    def _get_streams(self):
        # Construct manifest URL for this program.
        program_type, program_id = self.match.groups()
        manifest_type = self._program_type_map.get(program_type)
        if manifest_type is None:
            log.error(f"Unknown program type '{program_type}'")
            return None

        # Fetch program_id.
        res = self.session.http.get(self.url)
        m = self._program_id_re.search(res.text)
        if m is not None:
            program_id = m.group(1)
        elif program_id is None:
            log.error("Could not extract program ID from URL")
            return None

        manifest_url = urljoin(self._psapi_url, f"playback/manifest/{manifest_type}/{program_id}")

        # Extract media URL.
        res = self.session.http.get(manifest_url)
        manifest = self.session.http.json(res, schema=self._playable_schema)
        if 'nonPlayable' in manifest:
            reason = manifest["nonPlayable"]["reason"]
            log.error(f"Not playable ({reason})")
            return None
        self._set_metadata(manifest)
        asset = manifest['playable']['assets'][0]

        # Some streams such as podcasts are not HLS but plain files.
        if asset['format'] == 'HLS':
            return HLSStream.parse_variant_playlist(self.session, asset['url'])
        else:
            return [("live", HTTPStream(self.session, asset['url']))]

    def _set_metadata(self, manifest):
        luna = manifest.get("statistics").get("luna")
        if not luna:
            return
        data = luna["data"]
        self.category = data.get("category")
        self.title = data.get("title")


__plugin__ = NRK



from streamlink.plugins.app17 import App17
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlApp17(PluginCanHandleUrl):
    __plugin__ = App17

    should_match = [
        "https://17.live/en-US/live/123123"
        "https://17.live/en/live/123123",
        "https://17.live/ja/live/123123",
    ]



"""
$description Live TV channels and VODs from TVP, a Polish public, state-owned broadcaster.
$url tvp.info
$url tvp.pl
$type live, vod
$notes Some VODs may be geo-restricted. Authentication is not supported.
"""

import logging
import re
from typing import List, Optional, Tuple
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""https?://(?:www\.)?tvp\.info/"""), name="tvp_info")
@pluginmatcher(re.compile(
    r"""
        https?://
        (?:
            (?:tvpstream\.vod|stream)\.tvp\.pl(?:/(?:\?channel_id=(?P<channel_id>\d+))?)?$
            |
            vod\.tvp\.pl/[^/]+/.+,(?P<vod_id>\d+)$
        )
    """,
    re.VERBOSE,
))
class TVP(Plugin):
    _URL_PLAYER = "https://stream.tvp.pl/sess/TVPlayer2/embed.php"
    _URL_VOD = "https://vod.tvp.pl/api/products/{vod_id}/videos/playlist"

    def _get_video_id(self, channel_id: Optional[str]):
        items: List[Tuple[int, int]] = self.session.http.get(
            self.url,
            headers={
                # required, otherwise the next request for retrieving the HLS URL will be aborted by the server
                "Connection": "close",
            },
            schema=validate.Schema(
                validate.regex(re.compile(r"window\.__channels\s*=\s*(?P<json>\[.+?])\s*;", re.DOTALL)),
                validate.none_or_all(
                    validate.get("json"),
                    validate.parse_json(),
                    [
                        validate.all(
                            {
                                "id": int,
                                "items": validate.none_or_all(
                                    [{
                                        "video_id": int,
                                    }],
                                    validate.get((0, "video_id")),
                                ),
                            },
                            validate.union_get("id", "items"),
                        ),
                    ],
                ),
            ),
        )

        if channel_id is not None:
            _channel_id = int(channel_id)
            try:
                return next(item[1] for item in items if item[0] == _channel_id)
            except StopIteration:
                pass

        return items[0][1] if items else None

    def _get_live(self, channel_id: Optional[str]):
        video_id = self._get_video_id(channel_id)
        if not video_id:
            log.error("Could not find video ID")
            return

        log.debug(f"video ID: {video_id}")

        streams: Optional[List[Tuple[str, str]]] = self.session.http.get(
            self._URL_PLAYER,
            params={
                "ID": video_id,
                "autoPlay": "without_audio",
            },
            headers={
                "Referer": self.url,
            },
            schema=validate.Schema(
                validate.regex(re.compile(r"window\.__api__\s*=\s*(?P<json>\{.+?})\s*;", re.DOTALL)),
                validate.get("json"),
                validate.parse_json(),
                {
                    "result": validate.none_or_all(
                        {
                            "content": {
                                "files": [
                                    validate.all(
                                        {
                                            "type": str,
                                            "url": validate.url(),
                                        },
                                        validate.union_get("type", "url"),
                                    ),
                                ],
                            },
                        },
                        validate.get(("content", "files")),
                    ),
                },
                validate.get("result"),
            ),
        )
        if not streams:
            return

        def get(items, condition):
            return next((_url for _stype, _url in items if condition(_stype, urlparse(_url).path)), None)

        # prioritize HLSStream and get the first available stream
        url = get(streams, lambda t, _: t == "hls")
        if url:
            return HLSStream.parse_variant_playlist(self.session, url)

        # fall back to DASHStream
        url = get(streams, lambda t, p: t == "any_native" and p.endswith(".mpd"))
        if url:
            return DASHStream.parse_manifest(self.session, url)

        # fall back to HTTPStream
        url = get(streams, lambda t, p: t == "any_native" and p.endswith(".mp4"))
        if url:
            return {"vod": HTTPStream(self.session, url)}

    def _get_vod(self, vod_id):
        data = self.session.http.get(
            self._URL_VOD.format(vod_id=vod_id),
            params={
                "platform": "BROWSER",
                "videoType": "MOVIE",
            },
            acceptable_status=(200, 403),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {"code": "GEOIP_FILTER_FAILED"},
                    validate.all(
                        {
                            "sources": {
                                validate.optional("HLS"): [{
                                    "src": validate.url(),
                                }],
                            },
                        },
                        validate.get("sources"),
                    ),
                ),
            ),
        )

        if data.get("code") == "GEOIP_FILTER_FAILED":
            log.error("The content is not available in your region")
            return

        if data.get("HLS"):
            return HLSStream.parse_variant_playlist(self.session, data["HLS"][0]["src"])

    def _get_tvp_info_vod(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'window.__videoData')][1]/text()"),
                validate.none_or_all(
                    str,
                    validate.regex(re.compile(r"window\.__videoData\s*=\s*(?P<json>{.*?})\s*;", re.DOTALL)),
                    validate.get("json"),
                    validate.parse_json(),
                    {
                        "_id": int,
                        "title": str,
                    },
                ),
            ),
        )
        if not data:
            return

        self.id = data.get("_id")
        self.title = data.get("title")

        data = self.session.http.get(
            f"https://api.tvp.pl/tokenizer/token/{self.id}",
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "status": "OK",
                    "isGeoBlocked": bool,
                    "formats": [{
                        "mimeType": str,
                        "url": validate.url(),
                    }],
                },
            ),
        )
        log.debug(f"data={data}")

        if data.get("isGeoBlocked"):
            log.error("The content is not available in your region")
            return

        for format in data.get("formats"):
            if format.get("mimeType") == "application/x-mpegurl":
                return HLSStream.parse_variant_playlist(self.session, format.get("url"))

    def _get_streams(self):
        if self.matches["tvp_info"]:
            return self._get_tvp_info_vod()
        elif self.match["vod_id"]:
            return self._get_vod(self.match["vod_id"])
        else:
            return self._get_live(self.match["channel_id"])


__plugin__ = TVP



import argparse
import unittest
from unittest.mock import Mock, patch

import pytest

from streamlink.options import Argument, Arguments, Options
from streamlink.plugin import Plugin, pluginargument
from streamlink_cli.argparser import ArgumentParser
from streamlink_cli.main import setup_plugin_args, setup_plugin_options


class TestOptions(unittest.TestCase):
    def setUp(self):
        self.options = Options({
            "a_default": "default",
            "another-default": "default2"
        })

    def test_options(self):
        self.assertEqual(self.options.get("a_default"), "default")
        self.assertEqual(self.options.get("non_existing"), None)

        self.options.set("a_option", "option")
        self.assertEqual(self.options.get("a_option"), "option")

    def test_options_update(self):
        self.assertEqual(self.options.get("a_default"), "default")
        self.assertEqual(self.options.get("non_existing"), None)

        self.options.update({"a_option": "option"})
        self.assertEqual(self.options.get("a_option"), "option")

    def test_options_name_normalised(self):
        self.assertEqual(self.options.get("a_default"), "default")
        self.assertEqual(self.options.get("a-default"), "default")
        self.assertEqual(self.options.get("another-default"), "default2")
        self.assertEqual(self.options.get("another_default"), "default2")


class TestMappedOptions:
    class MappedOptions(Options):
        def _get_uppercase(self, key):
            return self.get_explicit(key.upper())

        def _get_add(self, key):
            return int(self.get_explicit(key)) + 1

        def _set_uppercase(self, key, value):
            self.set_explicit(key.upper(), value)

        def _set_add(self, key, value):
            self.set_explicit(key, int(value) + 1)

        _MAP_GETTERS = {
            "foo-bar": _get_uppercase,
            "baz": _get_add,
        }

        _MAP_SETTERS = {
            "foo-bar": _set_uppercase,
            "baz": _set_add,
        }

    @pytest.fixture
    def options(self):
        return self.MappedOptions({"foo-bar": 123, "baz": 100})

    def test_mapped_key(self, options: MappedOptions):
        assert options.get("foo-bar") is None
        assert options.get("foo_bar") is None
        assert options.get_explicit("foo-bar") == 123
        assert options.get_explicit("foo_bar") == 123
        assert options.get_explicit("FOO-BAR") is None
        assert options.get_explicit("FOO_BAR") is None

        options.set("foo-bar", 321)
        assert options.get("foo-bar") == 321
        assert options.get("foo_bar") == 321
        assert options.get_explicit("foo-bar") == 123
        assert options.get_explicit("foo_bar") == 123
        assert options.get_explicit("FOO-BAR") == 321
        assert options.get_explicit("FOO_BAR") == 321

    def test_mapped_value(self, options: MappedOptions):
        assert options.get("baz") == 101
        assert options.get_explicit("baz") == 100

        options.set("baz", 0)
        assert options.get("baz") == 2
        assert options.get_explicit("baz") == 1


class TestArgument(unittest.TestCase):
    def test_name(self):
        self.assertEqual(Argument("test-arg").argument_name("plugin"), "--plugin-test-arg")
        self.assertEqual(Argument("test-arg").namespace_dest("plugin"), "plugin_test_arg")
        self.assertEqual(Argument("test-arg").dest, "test_arg")

    def test_name_plugin(self):
        self.assertEqual(Argument("test-arg").argument_name("test_plugin"), "--test-plugin-test-arg")
        self.assertEqual(Argument("test-arg").namespace_dest("test_plugin"), "test_plugin_test_arg")
        self.assertEqual(Argument("test-arg").dest, "test_arg")

    def test_name_override(self):
        self.assertEqual(Argument("test", argument_name="override-name").argument_name("plugin"), "--override-name")
        self.assertEqual(Argument("test", argument_name="override-name").namespace_dest("plugin"), "override_name")
        self.assertEqual(Argument("test", argument_name="override-name").dest, "test")


class TestArguments(unittest.TestCase):
    def test_getter(self):
        test1 = Argument("test1")
        test2 = Argument("test2")
        args = Arguments(test1, test2)

        self.assertEqual(args.get("test1"), test1)
        self.assertEqual(args.get("test2"), test2)
        self.assertEqual(args.get("test3"), None)

    def test_iter(self):
        test1 = Argument("test1")
        test2 = Argument("test2")
        args = Arguments(test1, test2)

        i_args = iter(args)

        self.assertEqual(next(i_args), test1)
        self.assertEqual(next(i_args), test2)

    def test_requires(self):
        test1 = Argument("test1", requires="test2")
        test2 = Argument("test2", requires="test3")
        test3 = Argument("test3")

        args = Arguments(test1, test2, test3)

        self.assertEqual(list(args.requires("test1")), [test2, test3])

    def test_requires_invalid(self):
        test1 = Argument("test1", requires="test2")

        args = Arguments(test1)

        self.assertRaises(KeyError, lambda: list(args.requires("test1")))

    def test_requires_cycle(self):
        test1 = Argument("test1", requires="test2")
        test2 = Argument("test2", requires="test1")

        args = Arguments(test1, test2)

        self.assertRaises(RuntimeError, lambda: list(args.requires("test1")))

    def test_requires_cycle_deep(self):
        test1 = Argument("test1", requires="test-2")
        test2 = Argument("test-2", requires="test3")
        test3 = Argument("test3", requires="test1")

        args = Arguments(test1, test2, test3)

        self.assertRaises(RuntimeError, lambda: list(args.requires("test1")))

    def test_requires_cycle_self(self):
        test1 = Argument("test1", requires="test1")

        args = Arguments(test1)

        self.assertRaises(RuntimeError, lambda: list(args.requires("test1")))


class TestSetupOptions:
    def test_setup_plugin_args(self):
        session = Mock()
        plugin = Mock()
        parser = ArgumentParser(add_help=False)
        parser.add_argument("--global-arg1", default=123)
        parser.add_argument("--global-arg2", default=456)

        session.plugins = {"mock": plugin}
        plugin.arguments = Arguments(
            Argument("global-arg1", is_global=True),
            Argument("test1", default="default1"),
            Argument("test2", default="default2"),
            Argument("test3")
        )

        setup_plugin_args(session, parser)

        group_plugins = next((grp for grp in parser._action_groups if grp.title == "Plugin options"), None)  # pragma: no branch
        assert group_plugins is not None, "Adds the 'Plugin options' arguments group"
        assert group_plugins in parser.NESTED_ARGUMENT_GROUPS[None], "Adds the 'Plugin options' arguments group"
        group_plugin = next((grp for grp in parser._action_groups if grp.title == "Mock"), None)  # pragma: no branch
        assert group_plugin is not None, "Adds the 'Mock' arguments group"
        assert group_plugin in parser.NESTED_ARGUMENT_GROUPS[group_plugins], "Adds the 'Mock' arguments group"
        assert [item for action in group_plugin._group_actions for item in action.option_strings] \
            == ["--mock-test1", "--mock-test2", "--mock-test3"], \
            "Only adds plugin arguments and ignores global argument references"
        assert [item for action in parser._actions for item in action.option_strings] \
            == ["--global-arg1", "--global-arg2", "--mock-test1", "--mock-test2", "--mock-test3"], \
            "Parser has all arguments registered"

        assert plugin.options.get("global-arg1") == 123
        assert plugin.options.get("global-arg2") is None
        assert plugin.options.get("test1") == "default1"
        assert plugin.options.get("test2") == "default2"
        assert plugin.options.get("test3") is None

    def test_setup_plugin_options(self):
        @pluginargument("foo-foo", is_global=True)
        @pluginargument("bar-bar", default=456)
        @pluginargument("baz-baz", default=789, help=argparse.SUPPRESS)
        class FakePlugin(Plugin):
            def _get_streams(self):  # pragma: no cover
                pass

        session = Mock()
        parser = ArgumentParser()
        parser.add_argument("--foo-foo", default=123)

        session.plugins = {"plugin": FakePlugin}
        session.set_plugin_option = lambda name, key, value: session.plugins[name].options.update({key: value})

        with patch("streamlink_cli.main.args") as args:
            args.foo_foo = 321
            args.plugin_bar_bar = 654
            args.plugin_baz_baz = 987  # this wouldn't be set by the parser if the argument is suppressed

            setup_plugin_args(session, parser)
            assert FakePlugin.options.get("foo_foo") == 123, "Sets the global-argument's default value"
            assert FakePlugin.options.get("bar_bar") == 456, "Sets the plugin-argument's default value"
            assert FakePlugin.options.get("baz_baz") == 789, "Sets the suppressed plugin-argument's default value"

            setup_plugin_options(session, "plugin", FakePlugin)
            assert FakePlugin.options.get("foo_foo") == 321, "Sets the provided global-argument value"
            assert FakePlugin.options.get("bar_bar") == 654, "Sets the provided plugin-argument value"
            assert FakePlugin.options.get("baz_baz") == 789, "Doesn't set values of suppressed plugin-arguments"



"""Creates links that allows substituting the current release
within the title or target."""

import os.path

from docutils import nodes
from sphinx.util.nodes import split_explicit_title


def releaseref_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    config = inliner.document.settings.env.config
    text = text.replace("|version|", config.version)
    text = text.replace("|release|", config.release)

    has_explicit_title, title, target = split_explicit_title(text)
    if not has_explicit_title:
        title = os.path.basename(target)

    node = nodes.reference(rawtext, title, refuri=target, **options)

    return [node], []


def setup(app):
    app.add_role("releaseref", releaseref_role)



import unittest

from streamlink.plugins.bbciplayer import BBCiPlayer
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBBCiPlayer(PluginCanHandleUrl):
    __plugin__ = BBCiPlayer

    should_match = [
        "http://www.bbc.co.uk/iplayer/episode/b00ymh67/madagascar-1-island-of-marvels",
        "http://www.bbc.co.uk/iplayer/live/bbcone",
    ]

    should_not_match = [
        "http://www.bbc.co.uk/iplayer/",
    ]


class TestPluginBBCiPlayer(unittest.TestCase):
    def test_vpid_hash(self):
        self.assertEqual(
            "71c345435589c6ddeea70d6f252e2a52281ecbf3",
            BBCiPlayer._hash_vpid("1234567890")
        )



from threading import Lock, Thread
from unittest.mock import patch

import pytest

from tests.testutils.handshake import Handshake


class Producer(Thread):
    def __init__(self, exception=None):
        super().__init__(daemon=True, name="ProducerThread")
        self.handshake = Handshake()
        self.lock = Lock()
        self.value = 0
        self.exception = exception
        self.error = None

    def run(self):
        try:
            for _ in range(2):
                with self.handshake(self.exception):
                    with self.lock:
                        self.action()
        except Exception as err:
            self.error = err

    def action(self):
        self.value += 1


@pytest.fixture
def producer(request):
    thread = Producer(**getattr(request, "param", {}))
    yield thread
    thread.join(1)
    assert not thread.is_alive()


class TestSynchronization:
    def test_sync(self, producer: Producer):
        assert not producer.handshake.wait_ready(0), "Producer is not yet ready"
        assert not producer.handshake.wait_done(0), "Producer is not done"

        producer.start()
        assert producer.handshake.wait_ready(1), "Producer arrived in ready-state within at most 1 second"
        assert not producer.handshake.wait_done(0), "Producer is not in done-state"
        assert producer.value == 0

        assert producer.handshake.step(1), "Producer manages to execute one iteration within at most 1 second"
        assert not producer.handshake.wait_done(0), "Producer is not in done-state after completing one iteration"
        assert producer.value == 1
        assert producer.handshake.wait_ready(1), "Producer once again arrived in ready-state within at most 1 second"

        # make producer and consumer threads increment the value independently
        producer.handshake.go()
        with producer.lock:
            producer.value += 1
        assert producer.handshake.wait_done(1), "Producer is done after at most 1 second"
        assert producer.value == 3

        assert not producer.handshake.wait_done(0), "Producer is not in done-state anymore"
        assert not producer.handshake.wait_ready(0), "Producer's loop ended, not waiting in ready-state again"

    @pytest.mark.asyncio
    async def test_async(self, producer: Producer):
        assert not await producer.handshake.is_ready(0), "Producer is not yet ready"
        assert not await producer.handshake.is_done(0), "Producer is not done"

        producer.start()
        assert await producer.handshake.is_ready(1), "Producer arrived in ready-state within at most 1 second"
        assert not await producer.handshake.is_done(0), "Producer is not in done-state"
        assert producer.value == 0

        assert await producer.handshake.asyncstep(1), "Producer manages to execute one iteration within at most 1 second"
        assert not await producer.handshake.is_done(0), "Producer is not in done-state after completing one iteration"
        assert producer.value == 1
        assert await producer.handshake.is_ready(1), "Producer once again arrived in ready-state within at most 1 second"

        # make producer and consumer threads increment the value independently
        producer.handshake.go()
        is_done = producer.handshake.is_done(1)
        with producer.lock:
            producer.value += 1
        assert await is_done, "Producer is done after at most 1 second"
        assert producer.value == 3

        assert not await producer.handshake.is_done(0), "Producer is not in done-state anymore"
        assert not await producer.handshake.is_ready(0), "Producer's loop ended, not waiting in ready-state again"


class TestNoCaptureExceptions:
    def test_sync(self, producer: Producer):
        producer.start()

        # doesn't catch exception and doesn't raise in consumer thread
        with patch.object(producer, "action", side_effect=Exception):
            assert producer.handshake.step(1), "Doesn't catch any exceptions, doesn't raise in consumer thread"
        producer.join(1)
        assert not producer.is_alive(), "Producer thread has raised exception and has terminated"
        assert isinstance(producer.error, Exception)

    @pytest.mark.asyncio
    async def test_async(self, producer: Producer):
        producer.start()

        # doesn't catch exception and doesn't raise in consumer thread
        with patch.object(producer, "action", side_effect=Exception):
            assert await producer.handshake.asyncstep(1), "Doesn't catch any exceptions, doesn't raise in consumer thread"
        producer.join(1)
        assert not producer.is_alive(), "Producer thread has raised exception and has terminated"
        assert isinstance(producer.error, Exception)


class TestCaptureExceptions:
    @pytest.mark.parametrize("producer", [{"exception": TypeError}], indirect=True)
    def test_sync(self, producer: Producer):
        producer.start()

        # catches exception and raises in consumer thread
        with patch.object(producer, "action", side_effect=TypeError):
            with pytest.raises(TypeError):
                producer.handshake.step(1)

        assert not producer.handshake.wait_done(0), "Producer is not in done-state anymore after catching an exception"
        assert producer.handshake.wait_ready(1), "Producer once again arrived in ready-state within at most 1 second"

        with patch.object(producer, "action", side_effect=ValueError):
            assert producer.handshake.step(1), "Doesn't catch different exceptions, doesn't raise in consumer thread"
        producer.join(1)
        assert not producer.is_alive(), "Producer thread has raised exception and has terminated"
        assert isinstance(producer.error, ValueError)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("producer", [{"exception": TypeError}], indirect=True)
    async def test_async(self, producer: Producer):
        producer.start()

        # catches exception and raises in consumer thread
        with patch.object(producer, "action", side_effect=TypeError):
            with pytest.raises(TypeError):
                await producer.handshake.asyncstep(1)

        assert not await producer.handshake.is_done(0), "Producer is not in done-state anymore after catching an exception"
        assert await producer.handshake.is_ready(1), "Producer once again arrived in ready-state within at most 1 second"

        with patch.object(producer, "action", side_effect=ValueError):
            assert await producer.handshake.asyncstep(1), "Doesn't catch different exceptions, doesn't raise in consumer thread"
        producer.join(1)
        assert not producer.is_alive(), "Producer thread has raised exception and has terminated"
        assert isinstance(producer.error, ValueError)



"""
$description TV and live video game broadcasts, artist performances and personal daily-life video blogs & shows.
$url play.afreecatv.com
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter

log = logging.getLogger(__name__)


class AfreecaHLSStreamWriter(HLSStreamWriter):
    def should_filter_sequence(self, sequence):
        return "preloading" in sequence.segment.uri or super().should_filter_sequence(sequence)


class AfreecaHLSStreamReader(HLSStreamReader):
    __writer__ = AfreecaHLSStreamWriter


class AfreecaHLSStream(HLSStream):
    __reader__ = AfreecaHLSStreamReader


@pluginmatcher(re.compile(
    r"https?://play\.afreecatv\.com/(?P<username>\w+)(?:/(?P<bno>:\d+))?"
))
@pluginargument(
    "username",
    sensitive=True,
    requires=["password"],
    metavar="USERNAME",
    help="The username used to register with afreecatv.com.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="A afreecatv.com account password to use with --afreeca-username.",
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached AfreecaTV credentials to initiate a new session and reauthenticate.",
)
class AfreecaTV(Plugin):
    _re_bno = re.compile(r"var nBroadNo = (?P<bno>\d+);")

    CHANNEL_API_URL = "http://live.afreecatv.com/afreeca/player_live_api.php"
    CHANNEL_RESULT_OK = 1
    QUALITYS = ["original", "hd", "sd"]
    QUALITY_WEIGHTS = {
        "original": 1080,
        "hd": 720,
        "sd": 480,
    }

    _schema_channel = validate.Schema(
        {
            "CHANNEL": {
                "RESULT": validate.transform(int),
                validate.optional("BPWD"): str,
                validate.optional("BNO"): str,
                validate.optional("RMD"): str,
                validate.optional("AID"): str,
                validate.optional("CDN"): str,
            }
        },
        validate.get("CHANNEL")
    )
    _schema_stream = validate.Schema(
        {
            validate.optional("view_url"): validate.url(
                scheme=validate.any("rtmp", "http")
            ),
            "stream_status": str,
        }
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._authed = (
            self.session.http.cookies.get("PdboxBbs")
            and self.session.http.cookies.get("PdboxSaveTicket")
            and self.session.http.cookies.get("PdboxTicket")
            and self.session.http.cookies.get("PdboxUser")
            and self.session.http.cookies.get("RDB")
        )

    @classmethod
    def stream_weight(cls, key):
        weight = cls.QUALITY_WEIGHTS.get(key)
        if weight:
            return weight, "afreeca"

        return Plugin.stream_weight(key)

    def _get_channel_info(self, broadcast, username):
        data = {
            "bid": username,
            "bno": broadcast,
            "from_api": "0",
            "mode": "landing",
            "player_type": "html5",
            "pwd": "",
            "stream_type": "common",
            "type": "live",
        }
        res = self.session.http.post(self.CHANNEL_API_URL, data=data)
        return self.session.http.json(res, schema=self._schema_channel)

    def _get_hls_key(self, broadcast, username, quality):
        data = {
            "bid": username,
            "bno": broadcast,
            "from_api": "0",
            "mode": "landing",
            "player_type": "html5",
            "pwd": "",
            "quality": quality,
            "stream_type": "common",
            "type": "aid",
        }
        res = self.session.http.post(self.CHANNEL_API_URL, data=data)
        return self.session.http.json(res, schema=self._schema_channel)

    def _get_stream_info(self, broadcast, quality, rmd):
        params = {
            "return_type": "gs_cdn_pc_web",
            "broad_key": f"{broadcast}-common-{quality}-hls",
        }
        res = self.session.http.get(f"{rmd}/broad_stream_assign.html", params=params)
        return self.session.http.json(res, schema=self._schema_stream)

    def _get_hls_stream(self, broadcast, username, quality, rmd):
        keyjson = self._get_hls_key(broadcast, username, quality)

        if keyjson["RESULT"] != self.CHANNEL_RESULT_OK:
            return
        key = keyjson["AID"]

        info = self._get_stream_info(broadcast, quality, rmd)

        if "view_url" in info:
            return AfreecaHLSStream(self.session, info["view_url"], params={"aid": key})

    def _login(self, username, password):
        data = {
            "szWork": "login",
            "szType": "json",
            "szUid": username,
            "szPassword": password,
            "isSaveId": "true",
            "isSavePw": "false",
            "isSaveJoin": "false",
            "isLoginRetain": "Y",
        }
        res = self.session.http.post("https://login.afreecatv.com/app/LoginAction.php", data=data)
        data = self.session.http.json(res)
        log.trace(f"{data!r}")
        if data["RESULT"] != self.CHANNEL_RESULT_OK:
            return False
        self.save_cookies()
        return True

    def _get_streams(self):
        login_username = self.get_option("username")
        login_password = self.get_option("password")

        self.session.http.headers.update({"Referer": self.url, "Origin": "http://play.afreecatv.com"})

        if self.options.get("purge_credentials"):
            self.clear_cookies()
            self._authed = False
            log.info("All credentials were successfully removed")

        if self._authed:
            log.debug("Attempting to authenticate using cached cookies")
        elif login_username and login_password:
            log.debug("Attempting to login using username and password")
            if self._login(login_username, login_password):
                log.info("Login was successful")
            else:
                log.error("Failed to login")

        m = self.match.groupdict()
        username = m["username"]
        bno = m["bno"]
        if bno is None:
            res = self.session.http.get(self.url)
            m = self._re_bno.search(res.text)
            if not m:
                log.error("Could not find broadcast number.")
                return
            bno = m.group("bno")

        channel = self._get_channel_info(bno, username)
        log.trace(f"{channel!r}")
        if channel.get("BPWD") == "Y":
            log.error("Stream is Password-Protected")
            return
        elif channel.get("RESULT") == -6:
            log.error("Login required")
            return
        elif channel.get("RESULT") != self.CHANNEL_RESULT_OK:
            return

        (broadcast, rmd) = (channel["BNO"], channel["RMD"])
        if not (broadcast and rmd):
            return

        for qkey in self.QUALITYS:
            hls_stream = self._get_hls_stream(broadcast, username, qkey, rmd)
            if hls_stream:
                yield qkey, hls_stream


__plugin__ = AfreecaTV



from io import StringIO
from pathlib import PurePosixPath, PureWindowsPath
from time import time
from unittest.mock import Mock, call, patch

import freezegun
import pytest

from streamlink_cli.utils.progress import Progress, ProgressFormatter
from tests import posix_only, windows_only


class TestProgressFormatter:
    @pytest.fixture(scope="class")
    def params(self):
        return dict(
            written="WRITTEN",
            elapsed="ELAPSED",
            speed="SPEED",
            path=lambda *_: "PATH",
        )

    @pytest.mark.parametrize("term_width,expected", [
        (99, "[download] Written WRITTEN to PATH (ELAPSED @ SPEED)"),
        (63, "[download] Written WRITTEN to PATH (ELAPSED @ SPEED)"),
        (62, "[download] Written WRITTEN (ELAPSED @ SPEED)"),
        (44, "[download] Written WRITTEN (ELAPSED @ SPEED)"),
        (43, "[download] WRITTEN (ELAPSED @ SPEED)"),
        (36, "[download] WRITTEN (ELAPSED @ SPEED)"),
        (35, "[download] WRITTEN (ELAPSED)"),
        (28, "[download] WRITTEN (ELAPSED)"),
        (27, "[download] WRITTEN"),
        (1, "[download] WRITTEN"),
    ])
    def test_format(self, params, term_width, expected):
        with patch("streamlink_cli.utils.progress.ProgressFormatter.term_width", lambda: term_width):
            assert ProgressFormatter.format(ProgressFormatter.FORMATS, params) == expected

    @pytest.mark.parametrize("term_width,expected", [
        (99, "[download] Written WRITTEN to PATH (ELAPSED)"),
        (55, "[download] Written WRITTEN to PATH (ELAPSED)"),
        (54, "[download] Written WRITTEN (ELAPSED)"),
        (36, "[download] Written WRITTEN (ELAPSED)"),
        (35, "[download] WRITTEN (ELAPSED)"),
        (28, "[download] WRITTEN (ELAPSED)"),
        (27, "[download] WRITTEN"),
        (1, "[download] WRITTEN"),
    ])
    def test_format_nospeed(self, params, term_width, expected):
        with patch("streamlink_cli.utils.progress.ProgressFormatter.term_width", lambda: term_width):
            assert ProgressFormatter.format(ProgressFormatter.FORMATS_NOSPEED, params) == expected

    def test_format_missing(self, params):
        with patch("streamlink_cli.utils.progress.ProgressFormatter.term_width", lambda: 99):
            assert ProgressFormatter.format(ProgressFormatter.FORMATS, {"written": "0"}) == "[download] 0"

    def test_format_error(self, params):
        with patch("streamlink_cli.utils.progress.ProgressFormatter.term_width", lambda: 99):
            params = dict(**params)
            params["path"] = Mock(side_effect=ValueError("fail"))
            assert ProgressFormatter.format(ProgressFormatter.FORMATS, params) == "[download] Written WRITTEN (ELAPSED @ SPEED)"

    @pytest.mark.parametrize("size,expected", [
        (0, "0 bytes"),
        (2**10 - 1, "1023 bytes"),
        (2**10, "1.00 KiB"),
        (2**20 - 1, "1023.99 KiB"),
        (2**20, "1.00 MiB"),
        (2**30 - 1, "1023.99 MiB"),
        (2**30, "1.00 GiB"),
        (2**40 - 1, "1023.99 GiB"),
        (2**40, "1.00 TiB"),
        (2**50 - 1, "1023.99 TiB"),
        (2**50, "1024.00 TiB"),
    ])
    def test_format_filesize(self, size, expected):
        assert ProgressFormatter.format_filesize(size) == expected
        assert ProgressFormatter.format_filesize(float(size)) == expected
        assert ProgressFormatter.format_filesize(size, "/s") == f"{expected}/s"

    @pytest.mark.parametrize("elapsed,expected", [
        (-1, "0s"),
        (0, "0s"),
        (9, "9s"),
        (10, "10s"),
        (59, "59s"),
        (60, "1m00s"),
        (69, "1m09s"),
        (70, "1m10s"),
        (119, "1m59s"),
        (120, "2m00s"),
        (3599, "59m59s"),
        (3600, "1h00m00s"),
        (3659, "1h00m59s"),
        (3660, "1h01m00s"),
        (3661, "1h01m01s"),
        (86399, "23h59m59s"),
        (86400, "24h00m00s"),
        (172800, "48h00m00s"),
    ])
    def test_format_time(self, elapsed, expected):
        assert ProgressFormatter.format_time(elapsed) == expected

    _path_posix = PurePosixPath("/foobar/baz/some file name")
    _path_windows_abs = PureWindowsPath("C:\\foobar\\baz\\some file name")
    _path_windows_rel = PureWindowsPath("foobar\\baz\\some file name")
    _path_windows_unc = PureWindowsPath("\\\\?\\foobar\\baz\\some file name")

    @pytest.mark.parametrize("path,max_width,expected", [
        pytest.param(_path_posix, 26, "/foobar/baz/some file name", id="posix - full path"),
        pytest.param(_path_posix, 25, "…oobar/baz/some file name", id="posix - truncated by 1"),
        pytest.param(_path_posix, 24, "…obar/baz/some file name", id="posix - truncated by 2"),
        pytest.param(_path_posix, 23, "…bar/baz/some file name", id="posix - truncated by 3"),
        pytest.param(_path_posix, 22, "…ar/baz/some file name", id="posix - truncated by 4"),
        pytest.param(_path_posix, 21, "…r/baz/some file name", id="posix - truncated by 5"),
        pytest.param(_path_posix, 20, "…/baz/some file name", id="posix - truncated by 6"),
        pytest.param(_path_posix, 19, "…baz/some file name", id="posix - truncated by 7 (cuts off separator)"),
        pytest.param(_path_posix, 16, "…/some file name", id="posix - truncated (all parts except name)"),
        pytest.param(_path_posix, 15, "…some file name", id="posix - truncated (name without separator)"),
        pytest.param(_path_posix, 14, "…ome file name", id="posix - truncated name"),
        pytest.param(_path_windows_abs, 28, "C:\\foobar\\baz\\some file name", id="windows abs - full path"),
        pytest.param(_path_windows_abs, 27, "C:…oobar\\baz\\some file name", id="windows abs - truncated by 1"),
        pytest.param(_path_windows_abs, 26, "C:…obar\\baz\\some file name", id="windows abs - truncated by 2"),
        pytest.param(_path_windows_abs, 25, "C:…bar\\baz\\some file name", id="windows abs - truncated by 3"),
        pytest.param(_path_windows_abs, 24, "C:…ar\\baz\\some file name", id="windows abs - truncated by 4"),
        pytest.param(_path_windows_abs, 23, "C:…r\\baz\\some file name", id="windows abs - truncated by 5"),
        pytest.param(_path_windows_abs, 22, "C:…\\baz\\some file name", id="windows abs - truncated by 6"),
        pytest.param(_path_windows_abs, 21, "C:…baz\\some file name", id="windows abs - truncated by 7 (cuts off separator)"),
        pytest.param(_path_windows_abs, 18, "C:…\\some file name", id="windows abs - truncated (all parts except name)"),
        pytest.param(_path_windows_abs, 17, "C:…some file name", id="windows abs - truncated (name without separator)"),
        pytest.param(_path_windows_abs, 16, "C:…ome file name", id="windows abs - truncated name"),
        pytest.param(_path_windows_rel, 25, "foobar\\baz\\some file name", id="windows rel - full path"),
        pytest.param(_path_windows_rel, 24, "…obar\\baz\\some file name", id="windows rel - truncated by 1"),
        pytest.param(_path_windows_rel, 23, "…bar\\baz\\some file name", id="windows rel - truncated by 2"),
        pytest.param(_path_windows_rel, 22, "…ar\\baz\\some file name", id="windows rel - truncated by 3"),
        pytest.param(_path_windows_rel, 21, "…r\\baz\\some file name", id="windows rel - truncated by 4"),
        pytest.param(_path_windows_rel, 20, "…\\baz\\some file name", id="windows rel - truncated by 5"),
        pytest.param(_path_windows_rel, 19, "…baz\\some file name", id="windows rel - truncated by 6 (cuts off separator)"),
        pytest.param(_path_windows_rel, 16, "…\\some file name", id="windows rel - truncated (all parts except name)"),
        pytest.param(_path_windows_rel, 15, "…some file name", id="windows rel - truncated (name without separator)"),
        pytest.param(_path_windows_rel, 14, "…ome file name", id="windows rel - truncated name"),
        pytest.param(_path_windows_unc, 29, "\\\\?\\foobar\\baz\\some file name", id="windows UNC - full path"),
        pytest.param(_path_windows_unc, 28, "\\\\?\\…obar\\baz\\some file name", id="windows UNC - truncated by 1"),
        pytest.param(_path_windows_unc, 20, "\\\\?\\…\\some file name", id="windows UNC - truncated (all parts except name)"),
        pytest.param(_path_windows_unc, 19, "\\\\?\\…some file name", id="windows UNC - truncated (name without separator)"),
        pytest.param(_path_windows_unc, 18, "\\\\?\\…ome file name", id="windows UNC - truncated name"),
    ])
    def test_format_path(self, path, max_width, expected):
        with patch("os.path.sep", "\\" if type(path) is PureWindowsPath else "/"):
            assert ProgressFormatter.format_path(path, max_width) == expected


class TestWidth:
    @pytest.mark.parametrize("chars,expected", [
        ("ABCDEFGHIJ", 10),
        ("A你好世界こんにちは안녕하세요B", 30),
        ("·「」『』【】-=！@#￥%……&×（）", 30),
    ])
    def test_width(self, chars, expected):
        assert ProgressFormatter.width(chars) == expected

    @pytest.mark.parametrize("prefix,max_len,expected", [
        ("你好世界こんにちは안녕하세요CD", 10, "녕하세요CD"),
        ("你好世界こんにちは안녕하세요CD", 9, "하세요CD"),
        ("你好世界こんにちは안녕하세요CD", 23, "こんにちは안녕하세요CD"),
    ])
    def test_cut(self, prefix, max_len, expected):
        assert ProgressFormatter.cut(prefix, max_len) == expected


class TestPrint:
    @pytest.fixture(autouse=True)
    def _terminal_size(self):
        with patch("streamlink_cli.utils.progress.get_terminal_size") as mock_get_terminal_size:
            mock_get_terminal_size.return_value = Mock(columns=10)
            yield

    @pytest.fixture
    def stream(self):
        return StringIO()

    @pytest.fixture
    def progress(self, stream: StringIO):
        yield Progress(stream, Mock())

    @posix_only
    def test_print_posix(self, progress: Progress, stream: StringIO):
        progress.print_inplace("foo")
        progress.print_inplace("barbaz")
        progress.print_inplace("0123456789")
        progress.print_inplace("abcdefghijk")
        progress.print_end()
        assert stream.getvalue() == "\rfoo       \rbarbaz    \r0123456789\rabcdefghijk\n"

    @windows_only
    def test_print_windows(self, progress: Progress, stream: StringIO):
        progress.print_inplace("foo")
        progress.print_inplace("barbaz")
        progress.print_inplace("0123456789")
        progress.print_inplace("abcdefghijk")
        progress.print_end()
        assert stream.getvalue() == "\rfoo      \rbarbaz   \r0123456789\rabcdefghijk\n"


class TestProgress:
    def test_download_speed(self):
        kib = b"\x00" * 1024
        output_write = Mock()
        progress = Progress(
            Mock(write=output_write),
            PurePosixPath("../../the/path/where/we/write/to"),
            interval=1,
            history=3,
            threshold=2,
        )

        with freezegun.freeze_time("2000-01-01T00:00:00Z") as frozen_time, \
             patch("os.path.sep", "/"), \
             patch("streamlink_cli.utils.progress.ProgressFormatter.term_width", Mock(return_value=70)) as mock_width:
            progress.started = time()
            assert not output_write.call_args_list

            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 0 bytes to ../../the/path/where/we/write/to (0s)   ")

            frozen_time.tick()
            progress.write(kib * 1)
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 1.00 KiB to …th/where/we/write/to (1s @ 1.00 KiB/s)")

            frozen_time.tick()
            mock_width.return_value = 65
            progress.write(kib * 3)
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 4.00 KiB to …ere/we/write/to (2s @ 2.00 KiB/s)")

            frozen_time.tick()
            mock_width.return_value = 60
            progress.write(kib * 5)
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 9.00 KiB (3s @ 4.50 KiB/s)               ")

            frozen_time.tick()
            progress.write(kib * 7)
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 16.00 KiB (4s @ 7.50 KiB/s)              ")

            frozen_time.tick()
            progress.write(kib * 5)
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 21.00 KiB (5s @ 8.50 KiB/s)              ")

            frozen_time.tick()
            progress.update()
            assert output_write.call_args_list[-1] \
                   == call("\r[download] Written 21.00 KiB (6s @ 6.00 KiB/s)              ")



"""
$description 24-hour world, US and local news channel, based in the United States of America.
$url nbcnews.com
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?nbcnews\.com/now"
))
class NBCNews(Plugin):
    URL_API = "https://api-leap.nbcsports.com/feeds/assets/{}?application=NBCNews&format=nbc-player&platform=desktop"
    URL_TOKEN = "https://tokens.playmakerservices.com/"

    title = "NBC News Now"

    def _get_streams(self):
        self.id = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[@type='application/ld+json'][1]/text()"),
                validate.none_or_all(
                    validate.parse_json(),
                    {"embedUrl": validate.url()},
                    validate.get("embedUrl"),
                    validate.transform(lambda embed_url: embed_url.split("/")[-1]),
                ),
            ),
        )
        if self.id is None:
            return
        log.debug(f"API ID: {self.id}")

        stream = self.session.http.get(
            self.URL_API.format(self.id),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "videoSources": [{
                        "cdnSources": {
                            "primary": [{
                                "sourceUrl": validate.url(path=validate.endswith(".m3u8")),
                            }],
                        },
                    }],
                },
                validate.get(("videoSources", 0, "cdnSources", "primary", 0, "sourceUrl")),
            ),
        )

        url = self.session.http.post(
            self.URL_TOKEN,
            json={
                "requestorId": "nbcnews",
                "pid": self.id,
                "application": "NBCSports",
                "version": "v1",
                "platform": "desktop",
                "token": "",
                "resourceId": "",
                "inPath": "false",
                "authenticationType": "unauth",
                "cdn": "akamai",
                "url": stream,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "akamai": [{
                        "tokenizedUrl": validate.url(),
                    }],
                },
                validate.get(("akamai", 0, "tokenizedUrl")),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = NBCNews



from streamlink.plugins.trovo import Trovo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTrovo(PluginCanHandleUrl):
    __plugin__ = Trovo

    should_match_groups = [
        ("https://trovo.live/s/UserName", {"user": "UserName"}),
        ("https://trovo.live/s/UserName/abc", {"user": "UserName"}),
        ("https://trovo.live/s/UserName/123", {"user": "UserName"}),
        ("https://trovo.live/s/UserName/123?vid=vc-456&adtag=", {"user": "UserName", "video_id": "vc-456"}),
        ("https://trovo.live/s/UserName/123?vid=ltv-1_2_3&adtag=", {"user": "UserName", "video_id": "ltv-1_2_3"}),
        ("https://www.trovo.live/s/UserName", {"user": "UserName"}),
        ("https://www.trovo.live/s/UserName/abc", {"user": "UserName"}),
        ("https://www.trovo.live/s/UserName/123", {"user": "UserName"}),
        ("https://www.trovo.live/s/UserName/123?vid=vc-456&adtag=", {"user": "UserName", "video_id": "vc-456"}),
        ("https://www.trovo.live/s/UserName/123?vid=ltv-1_2_3&adtag=", {"user": "UserName", "video_id": "ltv-1_2_3"}),
    ]

    should_not_match = [
        "https://trovo.live/",
        "https://www.trovo.live/",
        "https://www.trovo.live/s/",
        "https://www.trovo.live/other/",
    ]



from typing import Any, Dict

from sphinx.addnodes import document
from sphinx.application import Sphinx


_CONFIG_VAR = "html_template_vars"


def update_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: document,
) -> None:
    for k, v in getattr(app.config, _CONFIG_VAR).items():
        context[k] = v


def setup(app: Sphinx) -> None:
    app.add_config_value(_CONFIG_VAR, {}, "")
    app.connect("html-page-context", update_context)



from streamlink.plugins.mediavitrina import MediaVitrina
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMediaVitrina(PluginCanHandleUrl):
    __plugin__ = MediaVitrina

    should_match = [
        "https://chetv.ru/online/",
        "https://ctc.ru/online/",
        "https://ctclove.ru/online/",
        "https://domashniy.ru/online",
        # player.mediavitrina.ru
        "https://player.mediavitrina.ru/5tv/moretv_web/player.html",
        "https://player.mediavitrina.ru/che/che_web/player.html",
        "https://player.mediavitrina.ru/ctc_ext/moretv_web/player.html",
        "https://player.mediavitrina.ru/ctc_love_ext/moretv_web/player.html",
        "https://player.mediavitrina.ru/ctc_love/ctclove_web/player.html",
        "https://player.mediavitrina.ru/ctc/ctcmedia_web/player.html?start=auto",
        "https://player.mediavitrina.ru/domashniy_ext/moretv_web/player.html",
        "https://player.mediavitrina.ru/domashniy/dom_web/player.html?start=auto",
        "https://player.mediavitrina.ru/gpm_tv3_v2/tv3/smotrim_web/611632488a33a/player.html",
        "https://player.mediavitrina.ru/iz/moretv_web/player.html",
        "https://player.mediavitrina.ru/kultura/limehd_web/player.html",
        "https://player.mediavitrina.ru/kultura/moretv_web/player.html",
        "https://player.mediavitrina.ru/mir/mir/moretv_web/player.html",
        "https://player.mediavitrina.ru/muztv/moretv_web/player.html",
        "https://player.mediavitrina.ru/rentv/moretv_web/player.html",
        "https://player.mediavitrina.ru/russia1/mailru_web/player.html",
        "https://player.mediavitrina.ru/russia1/moretv_web/player.html",
        "https://player.mediavitrina.ru/russia24/moretv_web/player.html",
        "https://player.mediavitrina.ru/russia24/vesti_ru_web/player.html?id",
        "https://player.mediavitrina.ru/spas/moretv_web/player.html",
        "https://player.mediavitrina.ru/tvc/tvc/moretv_web/player.html",
        "https://player.mediavitrina.ru/tvzvezda/moretv_web/player.html",
        "https://player.mediavitrina.ru/u_ott/u/moretv_web/player.html",
    ]

    should_not_match = [
        "https://1tv.ru/live",
        "https://ren.tv/live",
        "https://www.5-tv.ru/live/",
        "https://www.5-tv.ru/online/",
    ]



from streamlink.plugins.nos import NOS
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNOS(PluginCanHandleUrl):
    __plugin__ = NOS

    should_match = [
        'https://nos.nl/livestream/2220100-wk-sprint-schaatsen-1-000-meter-mannen.html',
        'https://nos.nl/collectie/13781/livestream/2385081-ek-voetbal-engeland-schotland',
        'https://nos.nl/collectie/13781/livestream/2385461-ek-voetbal-voorbeschouwing-italie-wales-18-00-uur',
        'https://nos.nl/collectie/13781/video/2385846-ek-in-2-21-gosens-show-tegen-portugal-en-weer-volle-bak-in-boedapest',
        'https://nos.nl/video/2385779-dronebeelden-tonen-spoor-van-vernieling-bij-leersum',
        'https://nos.nl/uitzendingen',
        'https://nos.nl/uitzendingen/livestream/2385462',
    ]

    should_not_match = [
        'https://nos.nl/artikel/2385784-explosieve-situatie-leidde-tot-verwoeste-huizen-en-omgewaaide-bomen-leersum',
    ]



"""
$description South Korean live-streaming platform for individual live streams.
$url pandalive.co.kr
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?pandalive\.co\.kr/"
))
class Pandalive(Plugin):
    def _get_streams(self):
        media_code = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"""routePath:\s*(?P<q>["'])(\\u002F|/)live(\\u002F|/)play(\\u002F|/)(?P<id>.+?)(?P=q)"""),
            validate.any(None, validate.get("id")),
        ))

        if not media_code:
            return

        log.debug(f"Media code: {media_code}")

        json = self.session.http.post(
            "https://api.pandalive.co.kr/v1/live/play",
            data={
                "action": "watch",
                "userId": media_code,
            },
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {
                        "media": {
                            "title": str,
                            "userId": str,
                            "userNick": str,
                            "isPw": bool,
                            "isLive": bool,
                            "liveType": str,
                        },
                        "PlayList": {
                            validate.optional("hls"): [{
                                "url": validate.url(),
                            }],
                            validate.optional("hls2"): [{
                                "url": validate.url(),
                            }],
                            validate.optional("hls3"): [{
                                "url": validate.url(),
                            }],
                        },
                        "result": bool,
                        "message": str,
                    },
                    {
                        "result": bool,
                        "message": str,
                    },
                ),
            ),
        )

        if not json["result"]:
            log.error(json["message"])
            return

        if not json["media"]["isLive"]:
            log.error("The broadcast has ended")
            return

        if json["media"]["isPw"]:
            log.error("The broadcast is password protected")
            return

        log.info(f"Broadcast type: {json['media']['liveType']}")

        self.author = f"{json['media']['userNick']} ({json['media']['userId']})"
        self.title = f"{json['media']['title']}"

        playlist = json["PlayList"]
        for key in ("hls", "hls2", "hls3"):
            # use the first available HLS stream
            if key in playlist and playlist[key]:
                # all stream qualities share the same URL, so just use the first one
                return HLSStream.parse_variant_playlist(self.session, playlist[key][0]["url"])


__plugin__ = Pandalive



from streamlink.plugins.tvibo import Tvibo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTvibo(PluginCanHandleUrl):
    __plugin__ = Tvibo

    should_match = [
        'http://player.tvibo.com/aztv/5929820',
        'http://player.tvibo.com/aztv/6858270/',
        'http://player.tvibo.com/aztv/3977238/',
    ]



from unittest.mock import Mock

import pytest

from streamlink.plugins.atresplayer import AtresPlayer
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAtresPlayer(PluginCanHandleUrl):
    __plugin__ = AtresPlayer

    should_match = [
        "http://www.atresplayer.com/directos/antena3/",
        "http://www.atresplayer.com/directos/lasexta/",
        "https://www.atresplayer.com/directos/antena3/",
        "https://www.atresplayer.com/flooxer/programas/unas/temporada-1/dario-eme-hache-sindy-takanashi-entrevista_123/",
    ]


class TestAtresPlayer:
    @pytest.mark.parametrize("url,expected", [
        ("http://www.atresplayer.com/directos/antena3", "https://www.atresplayer.com/directos/antena3/"),
        ("http://www.atresplayer.com/directos/antena3/", "https://www.atresplayer.com/directos/antena3/"),
        ("https://www.atresplayer.com/directos/antena3", "https://www.atresplayer.com/directos/antena3/"),
        ("https://www.atresplayer.com/directos/antena3/", "https://www.atresplayer.com/directos/antena3/"),
    ])
    def test_url(self, url, expected):
        plugin = AtresPlayer(Mock(), url)
        assert plugin.url == expected



from pathlib import Path
from typing import Optional

from streamlink.utils.formatter import Formatter as _BaseFormatter
from streamlink_cli.utils.path import replace_chars, replace_path


class Formatter(_BaseFormatter):
    title = _BaseFormatter.format

    def path(self, pathlike: str, charmap: Optional[str] = None) -> Path:
        def mapper(s):
            return replace_chars(s, charmap)

        def format_part(part):
            return self._format(part, mapper, {})

        return replace_path(pathlike, format_part)



"""
$description Live TV channels from RAI, an Italian public, state-owned broadcaster.
$url raiplay.it
$type live
$region Italy
"""

import logging
import re
from urllib.parse import parse_qsl, urlparse, urlunparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?raiplay\.it/dirette/(\w+)/?"
))
class RaiPlay(Plugin):
    _DEFAULT_MEDIAPOLIS_OUTPUT = "64"

    def _get_streams(self):
        json_url = self.session.http.get(self.url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//*[@data-video-json][1]/@data-video-json"),
        ))
        if not json_url:
            return

        json_url = urlunparse(urlparse(self.url)._replace(path=json_url))
        log.debug(f"Found JSON URL: {json_url}")

        content_url = self.session.http.get(json_url, schema=validate.Schema(
            validate.parse_json(),
            {"video": {"content_url": validate.url()}},
            validate.get(("video", "content_url")),
        ))
        if not content_url:
            log.error("Missing content URL")
            return

        content_url = content_url.replace("/relinkerServlet.mp4", "/relinkerServlet.htm")
        parsed = urlparse(content_url)
        params = dict(parse_qsl(parsed.query))
        if not parsed.path.endswith(".xml") and not params.get("output"):
            params["output"] = self._DEFAULT_MEDIAPOLIS_OUTPUT
            content_url = update_qsd(urlunparse(parsed), params)

        res = self.session.http.head(content_url)
        # status code will be 200 even if geo-blocked, so check the returned content-type
        if not res or not res.headers or res.headers["Content-Type"] == "video/mp4":
            log.error("Geo-restricted content")
            return

        stream_url = self.session.http.get(
            content_url,
            schema=validate.Schema(
                validate.parse_xml(),
                validate.xml_element(tag="Mediapolis"),
                validate.xml_xpath_string("./url[@type='content']/text()"),
                validate.none_or_all(
                    validate.transform(str.strip),
                    validate.url(path=validate.endswith(".m3u8")),
                ),
            ),
        )
        if not stream_url:
            log.error("Missing stream URL")
            return

        yield from HLSStream.parse_variant_playlist(self.session, stream_url).items()


__plugin__ = RaiPlay



from streamlink.plugins.mitele import Mitele
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMitele(PluginCanHandleUrl):
    __plugin__ = Mitele

    should_match = [
        "http://www.mitele.es/directo/bemad",
        "http://www.mitele.es/directo/boing",
        "http://www.mitele.es/directo/cuatro",
        "http://www.mitele.es/directo/divinity",
        "http://www.mitele.es/directo/energy",
        "http://www.mitele.es/directo/fdf",
        "http://www.mitele.es/directo/telecinco",
        "https://www.mitele.es/directo/gh-duo-24h-senal-1",
        "https://www.mitele.es/directo/gh-duo-24h-senal-2",
    ]

    should_not_match = [
        "http://www.mitele.es",
    ]



from streamlink.plugins.lrt import LRT
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlLRT(PluginCanHandleUrl):
    __plugin__ = LRT

    should_match = [
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-opus",
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-klasika",
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-radijas",
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-lituanica",
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-plius",
        "https://www.lrt.lt/mediateka/tiesiogiai/lrt-televizija",
    ]

    should_not_match = [
        "https://www.lrt.lt",
        "https://www.lrt.lt/mediateka/irasas/1013694276/savanoriai-tures-galimybe-pamatyti-popieziu-is-arciau",
    ]



from streamlink.stream.dash import DASHStream
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.hls import HLSStream, MuxedHLSStream
from streamlink.stream.http import HTTPStream
from streamlink.stream.stream import Stream, StreamIO
from streamlink.stream.wrappers import StreamIOIterWrapper, StreamIOThreadWrapper, StreamIOWrapper



import sys
from getpass import getpass
from json import dumps
from typing import Any, Dict, List, Optional, TextIO, Union

from streamlink.user_input import UserInputRequester
from streamlink_cli.utils import JSONEncoder


class ConsoleUserInputRequester(UserInputRequester):
    """
    Request input from the user on the console using the standard ask/askpass methods
    """
    def __init__(self, console):
        self.console = console

    def ask(self, prompt: str) -> str:
        if not sys.stdin.isatty():
            raise OSError("no TTY available")
        return self.console.ask(f"{prompt.strip()}: ")

    def ask_password(self, prompt: str) -> str:
        if not sys.stdin.isatty():
            raise OSError("no TTY available")
        return self.console.askpass(f"{prompt.strip()}: ")


class ConsoleOutput:
    def __init__(self, output: TextIO, json: bool = False):
        self.json = json
        self.output = output

    def ask(self, prompt: str) -> Optional[str]:
        if not sys.stdin.isatty():
            return None

        self.output.write(prompt)

        # noinspection PyBroadException
        try:
            return input().strip()
        except Exception:
            return None

    def askpass(self, prompt: str) -> Optional[str]:
        if not sys.stdin.isatty():
            return None

        return getpass(prompt, self.output)

    def msg(self, msg: str) -> None:
        if self.json:
            return
        self.output.write(f"{msg}\n")

    def msg_json(self, *objs: Any, **keywords: Any) -> None:
        if not self.json:
            return

        out: Union[List, Dict]
        if objs and isinstance(objs[0], list):
            out = []
            for obj in objs:
                if isinstance(obj, list):
                    out.extend(obj)
                else:
                    if hasattr(obj, "__json__") and callable(obj.__json__):
                        obj = obj.__json__()
                    out.append(obj)
            if keywords:
                out.append(keywords)
        else:
            out = {}
            for obj in objs:
                if hasattr(obj, "__json__") and callable(obj.__json__):
                    obj = obj.__json__()
                if not isinstance(obj, dict):
                    continue
                out.update(**obj)
            out.update(**keywords)

        msg = dumps(out, cls=JSONEncoder, indent=2)
        self.output.write(f"{msg}\n")

        if type(out) is dict and out.get("error"):
            sys.exit(1)

    def exit(self, msg: str) -> None:
        if self.json:
            self.msg_json(error=msg)
        else:
            self.msg(f"error: {msg}")

        sys.exit(1)


__all__ = ["ConsoleOutput", "ConsoleUserInputRequester"]



import datetime
import logging
import os
import sys
import unittest
from pathlib import Path, PosixPath, WindowsPath
from textwrap import dedent
from unittest.mock import Mock, call, patch

import freezegun
import pytest

import streamlink_cli.main
import tests.resources
from streamlink.exceptions import PluginError, StreamError
from streamlink.session import Streamlink
from streamlink.stream.stream import Stream
from streamlink_cli.compat import stdout
from streamlink_cli.main import (
    Formatter,
    NoPluginError,
    check_file_output,
    create_output,
    format_valid_streams,
    handle_stream,
    handle_url,
    output_stream,
    resolve_stream_name,
)
from streamlink_cli.output import FileOutput, PlayerOutput
from tests import posix_only, windows_only
from tests.plugin.testplugin import TestPlugin as _TestPlugin


# TODO: rewrite the entire mess


class FakePlugin(_TestPlugin):
    __module__ = "fake"
    _streams = {}  # type: ignore

    def streams(self, *args, **kwargs):
        return self._streams

    def _get_streams(self):  # pragma: no cover
        pass


class TestCLIMain(unittest.TestCase):
    def test_resolve_stream_name(self):
        a = Mock()
        b = Mock()
        c = Mock()
        d = Mock()
        e = Mock()
        streams = {
            "160p": a,
            "360p": b,
            "480p": c,
            "720p": d,
            "1080p": e,
            "worst": b,
            "best": d,
            "worst-unfiltered": a,
            "best-unfiltered": e
        }

        self.assertEqual(resolve_stream_name(streams, "unknown"), "unknown")
        self.assertEqual(resolve_stream_name(streams, "160p"), "160p")
        self.assertEqual(resolve_stream_name(streams, "360p"), "360p")
        self.assertEqual(resolve_stream_name(streams, "480p"), "480p")
        self.assertEqual(resolve_stream_name(streams, "720p"), "720p")
        self.assertEqual(resolve_stream_name(streams, "1080p"), "1080p")
        self.assertEqual(resolve_stream_name(streams, "worst"), "360p")
        self.assertEqual(resolve_stream_name(streams, "best"), "720p")
        self.assertEqual(resolve_stream_name(streams, "worst-unfiltered"), "160p")
        self.assertEqual(resolve_stream_name(streams, "best-unfiltered"), "1080p")

    def test_format_valid_streams(self):
        a = Mock()
        b = Mock()
        c = Mock()

        streams = {
            "audio": a,
            "720p": b,
            "1080p": c,
            "worst": b,
            "best": c
        }
        self.assertEqual(
            format_valid_streams(_TestPlugin, streams),
            ", ".join([
                "audio",
                "720p (worst)",
                "1080p (best)"
            ])
        )

        streams = {
            "audio": a,
            "720p": b,
            "1080p": c,
            "worst-unfiltered": b,
            "best-unfiltered": c
        }
        self.assertEqual(
            format_valid_streams(_TestPlugin, streams),
            ", ".join([
                "audio",
                "720p (worst-unfiltered)",
                "1080p (best-unfiltered)"
            ])
        )


class TestCLIMainHandleUrl:
    @pytest.mark.parametrize("side_effect,expected", [
        (NoPluginError("foo"), "No plugin can handle URL: fakeurl"),
        (PluginError("bar"), "bar"),
    ])
    def test_error(self, side_effect, expected):
        with patch("streamlink_cli.main.args", Mock(url="fakeurl")), \
             patch("streamlink_cli.main.streamlink", resolve_url=Mock(side_effect=side_effect)), \
             patch("streamlink_cli.main.console", exit=Mock(side_effect=SystemExit)) as mock_console:
            with pytest.raises(SystemExit):
                handle_url()
            assert mock_console.exit.mock_calls == [call(expected)]


class TestCLIMainJsonAndStreamUrl(unittest.TestCase):
    @patch("streamlink_cli.main.args", json=True, stream_url=True, subprocess_cmdline=False)
    @patch("streamlink_cli.main.console")
    def test_handle_stream_with_json_and_stream_url(self, console, args):
        stream = Mock()
        streams = dict(best=stream)

        plugin = FakePlugin(Mock(), "")
        plugin._streams = streams

        handle_stream(plugin, streams, "best")
        self.assertEqual(console.msg.mock_calls, [])
        self.assertEqual(console.msg_json.mock_calls, [call(
            stream,
            metadata=dict(
                id="test-id-1234-5678",
                author="Tѥst Āuƭhǿr",
                category=None,
                title="Test Title"
            )
        )])
        self.assertEqual(console.error.mock_calls, [])
        console.msg_json.mock_calls.clear()

        args.json = False
        handle_stream(plugin, streams, "best")
        self.assertEqual(console.msg.mock_calls, [call(stream.to_url())])
        self.assertEqual(console.msg_json.mock_calls, [])
        self.assertEqual(console.error.mock_calls, [])
        console.msg.mock_calls.clear()

        stream.to_url.side_effect = TypeError()
        handle_stream(plugin, streams, "best")
        self.assertEqual(console.msg.mock_calls, [])
        self.assertEqual(console.msg_json.mock_calls, [])
        self.assertEqual(console.exit.mock_calls, [call("The stream specified cannot be translated to a URL")])

    @patch("streamlink_cli.main.args", json=True, stream_url=True, stream=[], default_stream=[], retry_max=0, retry_streams=0)
    @patch("streamlink_cli.main.console")
    def test_handle_url_with_json_and_stream_url(self, console, args):
        stream = Mock()
        streams = dict(worst=Mock(), best=stream)

        class _FakePlugin(FakePlugin):
            __module__ = FakePlugin.__module__
            _streams = streams

        with patch("streamlink_cli.main.streamlink", resolve_url=Mock(return_value=("fake", _FakePlugin, ""))):
            handle_url()
            self.assertEqual(console.msg.mock_calls, [])
            self.assertEqual(console.msg_json.mock_calls, [call(
                plugin="fake",
                metadata=dict(
                    id="test-id-1234-5678",
                    author="Tѥst Āuƭhǿr",
                    category=None,
                    title="Test Title"
                ),
                streams=streams
            )])
            self.assertEqual(console.error.mock_calls, [])
            console.msg_json.mock_calls.clear()

            args.json = False
            handle_url()
            self.assertEqual(console.msg.mock_calls, [call(stream.to_manifest_url())])
            self.assertEqual(console.msg_json.mock_calls, [])
            self.assertEqual(console.error.mock_calls, [])
            console.msg.mock_calls.clear()

            stream.to_manifest_url.side_effect = TypeError()
            handle_url()
            self.assertEqual(console.msg.mock_calls, [])
            self.assertEqual(console.msg_json.mock_calls, [])
            self.assertEqual(console.exit.mock_calls, [call("The stream specified cannot be translated to a URL")])
            console.exit.mock_calls.clear()


class TestCLIMainCheckFileOutput(unittest.TestCase):
    @staticmethod
    def mock_path(path, is_file=True, resolve=""):
        return Mock(
            spec=Path(path),
            is_file=Mock(return_value=is_file),
            resolve=Mock(return_value=resolve),
            __str__=Mock(return_value=path)
        )

    @patch("streamlink_cli.main.log")
    def test_check_file_output(self, mock_log: Mock):
        path = self.mock_path("foo", is_file=False, resolve="/path/to/foo")
        output = check_file_output(path, False)
        self.assertIsInstance(output, FileOutput)
        self.assertIs(output.filename, path)
        self.assertEqual(mock_log.info.call_args_list, [call("Writing output to\n/path/to/foo")])
        self.assertEqual(mock_log.debug.call_args_list, [call("Checking file output")])

    def test_check_file_output_exists_force(self):
        path = self.mock_path("foo", is_file=True)
        output = check_file_output(path, True)
        self.assertIsInstance(output, FileOutput)
        self.assertIs(output.filename, path)

    @patch("streamlink_cli.main.console")
    @patch("streamlink_cli.main.sys")
    def test_check_file_output_exists_ask_yes(self, mock_sys: Mock, mock_console: Mock):
        mock_sys.stdin.isatty.return_value = True
        mock_console.ask = Mock(return_value="y")
        path = self.mock_path("foo", is_file=True)
        output = check_file_output(path, False)
        self.assertEqual(mock_console.ask.call_args_list, [call("File foo already exists! Overwrite it? [y/N] ")])
        self.assertIsInstance(output, FileOutput)
        self.assertIs(output.filename, path)

    @patch("streamlink_cli.main.console")
    @patch("streamlink_cli.main.sys")
    def test_check_file_output_exists_ask_no(self, mock_sys: Mock, mock_console: Mock):
        mock_sys.stdin.isatty.return_value = True
        mock_sys.exit.side_effect = SystemExit
        mock_console.ask = Mock(return_value="N")
        path = self.mock_path("foo", is_file=True)
        with self.assertRaises(SystemExit):
            check_file_output(path, False)
        self.assertEqual(mock_console.ask.call_args_list, [call("File foo already exists! Overwrite it? [y/N] ")])

    @patch("streamlink_cli.main.console")
    @patch("streamlink_cli.main.sys")
    def test_check_file_output_exists_ask_error(self, mock_sys: Mock, mock_console: Mock):
        mock_sys.stdin.isatty.return_value = True
        mock_sys.exit.side_effect = SystemExit
        mock_console.ask = Mock(return_value=None)
        path = self.mock_path("foo", is_file=True)
        with self.assertRaises(SystemExit):
            check_file_output(path, False)
        self.assertEqual(mock_console.ask.call_args_list, [call("File foo already exists! Overwrite it? [y/N] ")])

    @patch("streamlink_cli.main.console")
    @patch("streamlink_cli.main.sys")
    def test_check_file_output_exists_notty(self, mock_sys: Mock, mock_console: Mock):
        mock_sys.stdin.isatty.return_value = False
        mock_sys.exit.side_effect = SystemExit
        path = self.mock_path("foo", is_file=True)
        with self.assertRaises(SystemExit):
            check_file_output(path, False)
        self.assertEqual(mock_console.ask.call_args_list, [])


# TODO: don't use Mock() for mocking args, use a custom argparse.Namespace instead
class TestCLIMainCreateOutput(unittest.TestCase):
    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.console", Mock())
    @patch("streamlink_cli.main.DEFAULT_STREAM_METADATA", {"title": "bar"})
    def test_create_output_no_file_output_options(self, args: Mock):
        formatter = Formatter({
            "author": lambda: "foo"
        })
        args.output = None
        args.stdout = None
        args.record = None
        args.record_and_pipe = None
        args.player_fifo = False
        args.player_http = False
        args.title = None
        args.url = "URL"
        args.player = "mpv"
        args.player_args = ""

        output = create_output(formatter)
        assert type(output) is PlayerOutput
        assert output.title == "URL"

        args.title = "{author} - {title}"
        output = create_output(formatter)
        assert type(output) is PlayerOutput
        assert output.title == "foo - bar"

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.check_file_output")
    def test_create_output_file_output(self, mock_check_file_output: Mock, args: Mock):
        formatter = Formatter({})
        mock_check_file_output.side_effect = lambda path, force: FileOutput(path)
        args.output = "foo"
        args.stdout = None
        args.record = None
        args.record_and_pipe = None
        args.force = False
        args.fs_safe_rules = None

        output = create_output(formatter)
        assert mock_check_file_output.call_args_list == [call(Path("foo"), False)]
        assert type(output) is FileOutput
        assert output.filename == Path("foo")
        assert output.fd is None
        assert output.record is None

    @patch("streamlink_cli.main.args")
    def test_create_output_stdout(self, args: Mock):
        formatter = Formatter({})
        args.output = None
        args.stdout = True
        args.record = None
        args.record_and_pipe = None

        output = create_output(formatter)
        assert type(output) is FileOutput
        assert output.filename is None
        assert output.fd is stdout
        assert output.record is None

        args.output = "-"
        args.stdout = False
        output = create_output(formatter)
        assert type(output) is FileOutput
        assert output.filename is None
        assert output.fd is stdout
        assert output.record is None

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.check_file_output")
    def test_create_output_record_and_pipe(self, mock_check_file_output: Mock, args: Mock):
        formatter = Formatter({})
        mock_check_file_output.side_effect = lambda path, force: FileOutput(path)
        args.output = None
        args.stdout = None
        args.record_and_pipe = "foo"
        args.force = False
        args.fs_safe_rules = None

        output = create_output(formatter)
        assert mock_check_file_output.call_args_list == [call(Path("foo"), False)]
        assert type(output) is FileOutput
        assert output.filename is None
        assert output.fd is stdout
        assert type(output.record) is FileOutput
        assert output.record.filename == Path("foo")
        assert output.record.fd is None
        assert output.record.record is None

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.check_file_output")
    @patch("streamlink_cli.main.DEFAULT_STREAM_METADATA", {"title": "bar"})
    def test_create_output_record(self, mock_check_file_output: Mock, args: Mock):
        formatter = Formatter({
            "author": lambda: "foo"
        })
        mock_check_file_output.side_effect = lambda path, force: FileOutput(path)
        args.output = None
        args.stdout = None
        args.record = "foo"
        args.record_and_pipe = None
        args.force = False
        args.fs_safe_rules = None
        args.title = None
        args.url = "URL"
        args.player = "mpv"
        args.player_args = ""
        args.player_fifo = None
        args.player_http = None

        output = create_output(formatter)
        assert type(output) is PlayerOutput
        assert output.title == "URL"
        assert type(output.record) is FileOutput
        assert output.record.filename == Path("foo")
        assert output.record.fd is None
        assert output.record.record is None

        args.title = "{author} - {title}"
        output = create_output(formatter)
        assert type(output) is PlayerOutput
        assert output.title == "foo - bar"
        assert type(output.record) is FileOutput
        assert output.record.filename == Path("foo")
        assert output.record.fd is None
        assert output.record.record is None

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.DEFAULT_STREAM_METADATA", {"title": "bar"})
    def test_create_output_record_stdout(self, args: Mock):
        formatter = Formatter({
            "author": lambda: "foo"
        })
        args.output = None
        args.stdout = None
        args.record = "-"
        args.record_and_pipe = None
        args.force = False
        args.fs_safe_rules = None
        args.title = "{author} - {title}"
        args.url = "URL"
        args.player = "mpv"
        args.player_args = ""
        args.player_fifo = None
        args.player_http = None

        output = create_output(formatter)
        assert type(output) is PlayerOutput
        assert output.title == "foo - bar"
        assert type(output.record) is FileOutput
        assert output.record.filename is None
        assert output.record.fd is stdout
        assert output.record.record is None

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.console")
    def test_create_output_record_and_other_file_output(self, console: Mock, args: Mock):
        formatter = Formatter({})
        args.output = None
        args.stdout = True
        args.record_and_pipe = True
        create_output(formatter)
        console.exit.assert_called_with("Cannot use record options with other file output options.")

    @patch("streamlink_cli.main.args")
    @patch("streamlink_cli.main.console")
    def test_create_output_no_default_player(self, console: Mock, args: Mock):
        formatter = Formatter({})
        args.output = None
        args.stdout = False
        args.record_and_pipe = False
        args.player = None
        console.exit.side_effect = SystemExit
        with self.assertRaises(SystemExit):
            create_output(formatter)
        self.assertRegex(
            console.exit.call_args_list[0][0][0],
            r"^The default player \(\w+\) does not seem to be installed\."
        )


class TestCLIMainHandleStream(unittest.TestCase):
    @patch("streamlink_cli.main.output_stream")
    @patch("streamlink_cli.main.args")
    def test_handle_stream_output_stream(self, args: Mock, mock_output_stream: Mock):
        args.json = False
        args.subprocess_cmdline = False
        args.stream_url = False
        args.output = False
        args.stdout = False
        args.player_passthrough = []
        args.player_external_http = False
        args.player_continuous_http = False
        mock_output_stream.return_value = True

        session = Mock()
        plugin = FakePlugin(session, "")
        stream = Stream(session)
        streams = {"best": stream}

        handle_stream(plugin, streams, "best")
        self.assertEqual(mock_output_stream.call_count, 1)
        paramStream, paramFormatter = mock_output_stream.call_args[0]
        self.assertIs(paramStream, stream)
        self.assertIsInstance(paramFormatter, Formatter)


class TestCLIMainOutputStream(unittest.TestCase):
    @patch("streamlink_cli.main.args", Mock(retry_open=2))
    @patch("streamlink_cli.main.log")
    @patch("streamlink_cli.main.console")
    def test_stream_failure_no_output_open(self, mock_console: Mock, mock_log: Mock):
        output = Mock()
        stream = Mock(
            __str__=lambda _: "fake-stream",
            open=Mock(side_effect=StreamError("failure"))
        )
        formatter = Formatter({})

        with patch("streamlink_cli.main.output", Mock()), \
             patch("streamlink_cli.main.create_output", return_value=output):
            output_stream(stream, formatter)

        self.assertEqual(mock_log.error.call_args_list, [
            call("Try 1/2: Could not open stream fake-stream (Could not open stream: failure)"),
            call("Try 2/2: Could not open stream fake-stream (Could not open stream: failure)"),
        ])
        self.assertEqual(mock_console.exit.call_args_list, [
            call("Could not open stream fake-stream, tried 2 times, exiting")
        ])
        self.assertFalse(output.open.called, "Does not open the output on stream error")


class _TestCLIMainLogging(unittest.TestCase):
    # stop test execution at the setup_signals() call, as we're not interested in what comes afterwards
    class StopTest(Exception):
        pass

    @classmethod
    def subject(cls, argv, **kwargs):
        session = Streamlink()
        session.load_plugins(str(Path(tests.__path__[0]) / "plugin"))

        with patch("streamlink_cli.main.os.geteuid", create=True, new=Mock(return_value=kwargs.get("euid", 1000))), \
             patch("streamlink_cli.main.streamlink", session), \
             patch("streamlink_cli.main.setup_signals", side_effect=cls.StopTest), \
             patch("streamlink_cli.main.CONFIG_FILES", []), \
             patch("streamlink_cli.main.setup_streamlink"), \
             patch("streamlink_cli.main.setup_plugins"), \
             patch("streamlink.session.Streamlink.load_builtin_plugins"), \
             patch("sys.argv") as mock_argv:
            mock_argv.__getitem__.side_effect = lambda x: argv[x]
            try:
                streamlink_cli.main.main()
            except cls.StopTest:
                pass

    def tearDown(self):
        streamlink_cli.main.logger.root.handlers.clear()

    # python >=3.7.2: https://bugs.python.org/issue35046
    _write_call_log_cli_info = (
        [call("[cli][info] foo\n")]
        if sys.version_info >= (3, 7, 2) else
        [call("[cli][info] foo"), call("\n")]
    )
    _write_call_console_msg = [call("bar\n")]
    _write_call_console_msg_error = [call("error: bar\n")]
    _write_call_console_msg_json = [call("{\n  \"error\": \"bar\"\n}\n")]

    _write_calls = _write_call_log_cli_info + _write_call_console_msg

    def write_file_and_assert(self, mock_mkdir: Mock, mock_write: Mock, mock_stdout: Mock):
        streamlink_cli.main.log.info("foo")
        streamlink_cli.main.console.msg("bar")
        self.assertEqual(mock_mkdir.mock_calls, [call(parents=True, exist_ok=True)])
        self.assertEqual(mock_write.mock_calls, self._write_calls)
        self.assertFalse(mock_stdout.write.called)


class TestCLIMainLoggingStreams(_TestCLIMainLogging):
    # python >=3.7.2: https://bugs.python.org/issue35046
    _write_call_log_testcli_err = (
        [call("[test_cli_main][error] baz\n")]
        if sys.version_info >= (3, 7, 2) else
        [call("[test_cli_main][error] baz"), call("\n")]
    )

    def subject(self, argv, stream=None):
        super().subject(argv)
        childlogger = logging.getLogger("streamlink.test_cli_main")

        with self.assertRaises(SystemExit):
            streamlink_cli.main.log.info("foo")
            childlogger.error("baz")
            streamlink_cli.main.console.exit("bar")

        self.assertIs(streamlink_cli.main.log.parent.handlers[0].stream, stream)
        self.assertIs(childlogger.parent.handlers[0].stream, stream)
        self.assertIs(streamlink_cli.main.console.output, stream)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_stdout(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--stdout"], mock_stderr)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_output_eq_file(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--output=foo"], mock_stdout)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_output_eq_dash(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--output=-"], mock_stderr)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_record_eq_file(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--record=foo"], mock_stdout)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_record_eq_dash(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--record=-"], mock_stderr)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_stream_record_and_pipe(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--record-and-pipe=foo"], mock_stderr)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_no_pipe_no_json(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink"], mock_stdout)
        self.assertEqual(mock_stdout.write.mock_calls,
                         self._write_call_log_cli_info + self._write_call_log_testcli_err + self._write_call_console_msg_error)
        self.assertEqual(mock_stderr.write.mock_calls, [])

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_no_pipe_json(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--json"], mock_stdout)
        self.assertEqual(mock_stdout.write.mock_calls, self._write_call_console_msg_json)
        self.assertEqual(mock_stderr.write.mock_calls, [])

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_pipe_no_json(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--stdout"], mock_stderr)
        self.assertEqual(mock_stdout.write.mock_calls, [])
        self.assertEqual(mock_stderr.write.mock_calls,
                         self._write_call_log_cli_info + self._write_call_log_testcli_err + self._write_call_console_msg_error)

    @patch("sys.stderr")
    @patch("sys.stdout")
    def test_pipe_json(self, mock_stdout: Mock, mock_stderr: Mock):
        self.subject(["streamlink", "--stdout", "--json"], mock_stderr)
        self.assertEqual(mock_stdout.write.mock_calls, [])
        self.assertEqual(mock_stderr.write.mock_calls, self._write_call_console_msg_json)


class TestCLIMainLoggingInfos(_TestCLIMainLogging):
    @posix_only
    @patch("streamlink_cli.main.log")
    def test_log_root_warning(self, mock_log):
        self.subject(["streamlink"], euid=0)
        self.assertEqual(mock_log.info.mock_calls, [call("streamlink is running as root! Be careful!")])

    @patch("streamlink_cli.main.log")
    @patch("streamlink_cli.main.streamlink_version", "streamlink")
    @patch("streamlink_cli.main.importlib_metadata")
    @patch("streamlink_cli.main.log_current_arguments", Mock(side_effect=_TestCLIMainLogging.StopTest))
    @patch("platform.python_version", Mock(return_value="python"))
    def test_log_current_versions(self, mock_importlib_metadata: Mock, mock_log: Mock):
        class FakePackageNotFoundError(Exception):
            pass

        def version(dist):
            if dist == "foo":
                return "1.2.3"
            if dist == "bar-baz":
                return "2.0.0"
            raise FakePackageNotFoundError()

        mock_importlib_metadata.PackageNotFoundError = FakePackageNotFoundError
        mock_importlib_metadata.requires.return_value = ["foo>1", "bar-baz==2", "qux~=3"]
        mock_importlib_metadata.version.side_effect = version

        self.subject(["streamlink", "--loglevel", "info"])
        self.assertEqual(mock_log.debug.mock_calls, [], "Doesn't log anything if not debug logging")

        with patch("sys.platform", "linux"), \
             patch("platform.platform", Mock(return_value="linux")):
            self.subject(["streamlink", "--loglevel", "debug"])
            assert mock_importlib_metadata.requires.mock_calls == [call("streamlink")]
            assert mock_log.debug.mock_calls == [
                call("OS:         linux"),
                call("Python:     python"),
                call("Streamlink: streamlink"),
                call("Dependencies:"),
                call(" foo: 1.2.3"),
                call(" bar-baz: 2.0.0"),
            ]
            mock_importlib_metadata.requires.reset_mock()
            mock_log.debug.reset_mock()

        with patch("sys.platform", "darwin"), \
             patch("platform.mac_ver", Mock(return_value=["0.0.0"])):
            self.subject(["streamlink", "--loglevel", "debug"])
            assert mock_importlib_metadata.requires.mock_calls == [call("streamlink")]
            assert mock_log.debug.mock_calls == [
                call("OS:         macOS 0.0.0"),
                call("Python:     python"),
                call("Streamlink: streamlink"),
                call("Dependencies:"),
                call(" foo: 1.2.3"),
                call(" bar-baz: 2.0.0"),
            ]
            mock_importlib_metadata.requires.reset_mock()
            mock_log.debug.reset_mock()

        with patch("sys.platform", "win32"), \
             patch("platform.system", Mock(return_value="Windows")), \
             patch("platform.release", Mock(return_value="0.0.0")):
            self.subject(["streamlink", "--loglevel", "debug"])
            assert mock_importlib_metadata.requires.mock_calls == [call("streamlink")]
            assert mock_log.debug.mock_calls == [
                call("OS:         Windows 0.0.0"),
                call("Python:     python"),
                call("Streamlink: streamlink"),
                call("Dependencies:"),
                call(" foo: 1.2.3"),
                call(" bar-baz: 2.0.0"),
            ]
            mock_importlib_metadata.requires.reset_mock()
            mock_log.debug.reset_mock()

    @patch("streamlink_cli.main.log")
    def test_log_current_arguments(self, mock_log):
        self.subject([
            "streamlink",
            "--loglevel", "info"
        ])
        self.assertEqual(mock_log.debug.mock_calls, [], "Doesn't log anything if not debug logging")

        self.subject([
            "streamlink",
            "--loglevel", "debug",
            "-p", "custom",
            "--testplugin-bool",
            "--testplugin-password=secret",
            "test.se/channel",
            "best,worst"
        ])
        self.assertEqual(
            mock_log.debug.mock_calls[-7:],
            [
                call("Arguments:"),
                call(" url=test.se/channel"),
                call(" stream=['best', 'worst']"),
                call(" --loglevel=debug"),
                call(" --player=custom"),
                call(" --testplugin-bool=True"),
                call(" --testplugin-password=********")
            ]
        )


class TestCLIMainLoggingLogfile(_TestCLIMainLogging):
    @patch("sys.stdout")
    @patch("builtins.open")
    def test_logfile_no_logfile(self, mock_open, mock_stdout):
        self.subject(["streamlink"])
        streamlink_cli.main.log.info("foo")
        streamlink_cli.main.console.msg("bar")
        self.assertEqual(streamlink_cli.main.console.output, sys.stdout)
        self.assertFalse(mock_open.called)
        self.assertEqual(mock_stdout.write.mock_calls, self._write_calls)

    @patch("sys.stdout")
    @patch("builtins.open")
    def test_logfile_loglevel_none(self, mock_open, mock_stdout):
        self.subject(["streamlink", "--loglevel", "none", "--logfile", "foo"])
        streamlink_cli.main.log.info("foo")
        streamlink_cli.main.console.msg("bar")
        self.assertEqual(streamlink_cli.main.console.output, sys.stdout)
        self.assertFalse(mock_open.called)
        self.assertEqual(mock_stdout.write.mock_calls, [call("bar\n")])

    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    def test_logfile_path_relative(self, mock_open, mock_stdout):
        path = Path("foo").resolve()
        self.subject(["streamlink", "--logfile", "foo"])
        self.write_file_and_assert(
            mock_mkdir=path.mkdir,
            mock_write=mock_open(str(path), "a").write,
            mock_stdout=mock_stdout
        )


@posix_only
class TestCLIMainLoggingLogfilePosix(_TestCLIMainLogging):
    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    def test_logfile_path_absolute(self, mock_open, mock_stdout):
        self.subject(["streamlink", "--logfile", "/foo/bar"])
        self.write_file_and_assert(
            mock_mkdir=PosixPath("/foo").mkdir,
            mock_write=mock_open("/foo/bar", "a").write,
            mock_stdout=mock_stdout
        )

    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    def test_logfile_path_expanduser(self, mock_open, mock_stdout):
        with patch.dict(os.environ, {"HOME": "/foo"}):
            self.subject(["streamlink", "--logfile", "~/bar"])
        self.write_file_and_assert(
            mock_mkdir=PosixPath("/foo").mkdir,
            mock_write=mock_open("/foo/bar", "a").write,
            mock_stdout=mock_stdout
        )

    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    @freezegun.freeze_time(datetime.datetime(2000, 1, 2, 3, 4, 5))
    def test_logfile_path_auto(self, mock_open, mock_stdout):
        with patch("streamlink_cli.constants.LOG_DIR", PosixPath("/foo")):
            self.subject(["streamlink", "--logfile", "-"])
        self.write_file_and_assert(
            mock_mkdir=PosixPath("/foo").mkdir,
            mock_write=mock_open("/foo/2000-01-02_03-04-05.log", "a").write,
            mock_stdout=mock_stdout
        )


@windows_only
class TestCLIMainLoggingLogfileWindows(_TestCLIMainLogging):
    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    def test_logfile_path_absolute(self, mock_open, mock_stdout):
        self.subject(["streamlink", "--logfile", "C:\\foo\\bar"])
        self.write_file_and_assert(
            mock_mkdir=WindowsPath("C:\\foo").mkdir,
            mock_write=mock_open("C:\\foo\\bar", "a").write,
            mock_stdout=mock_stdout
        )

    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    def test_logfile_path_expanduser(self, mock_open, mock_stdout):
        with patch.dict(os.environ, {"USERPROFILE": "C:\\foo"}):
            self.subject(["streamlink", "--logfile", "~\\bar"])
        self.write_file_and_assert(
            mock_mkdir=WindowsPath("C:\\foo").mkdir,
            mock_write=mock_open("C:\\foo\\bar", "a").write,
            mock_stdout=mock_stdout
        )

    @patch("sys.stdout")
    @patch("builtins.open")
    @patch("pathlib.Path.mkdir", Mock())
    @freezegun.freeze_time(datetime.datetime(2000, 1, 2, 3, 4, 5))
    def test_logfile_path_auto(self, mock_open, mock_stdout):
        with patch("streamlink_cli.constants.LOG_DIR", WindowsPath("C:\\foo")):
            self.subject(["streamlink", "--logfile", "-"])
        self.write_file_and_assert(
            mock_mkdir=WindowsPath("C:\\foo").mkdir,
            mock_write=mock_open("C:\\foo\\2000-01-02_03-04-05.log", "a").write,
            mock_stdout=mock_stdout
        )


class TestCLIMainPrint(unittest.TestCase):
    def subject(self):
        with patch.object(Streamlink, "load_builtin_plugins"), \
             patch.object(Streamlink, "resolve_url") as mock_resolve_url, \
             patch.object(Streamlink, "resolve_url_no_redirect") as mock_resolve_url_no_redirect:
            session = Streamlink()
            session.load_plugins(str(Path(tests.__path__[0]) / "plugin"))
            with patch("streamlink_cli.main.os.geteuid", create=True, new=Mock(return_value=1000)), \
                 patch("streamlink_cli.main.streamlink", session), \
                 patch("streamlink_cli.main.CONFIG_FILES", []), \
                 patch("streamlink_cli.main.setup_streamlink"), \
                 patch("streamlink_cli.main.setup_plugins"), \
                 patch("streamlink_cli.main.setup_signals"):
                with self.assertRaises(SystemExit) as cm:
                    streamlink_cli.main.main()
                self.assertEqual(cm.exception.code, 0)
                mock_resolve_url.assert_not_called()
                mock_resolve_url_no_redirect.assert_not_called()

    @staticmethod
    def get_stdout(mock_stdout):
        return "".join([call_arg[0][0] for call_arg in mock_stdout.write.call_args_list])

    @patch("sys.stdout")
    @patch("sys.argv", ["streamlink"])
    def test_print_usage(self, mock_stdout):
        self.subject()
        self.assertEqual(
            self.get_stdout(mock_stdout),
            "usage: streamlink [OPTIONS] <URL> [STREAM]\n\n"
            + "Use -h/--help to see the available options or read the manual at https://streamlink.github.io\n"
        )

    @patch("sys.stdout")
    @patch("sys.argv", ["streamlink", "--help"])
    def test_print_help(self, mock_stdout):
        self.subject()
        output = self.get_stdout(mock_stdout)
        self.assertIn(
            "usage: streamlink [OPTIONS] <URL> [STREAM]",
            output
        )
        self.assertIn(
            dedent("""
                Streamlink is a command-line utility that extracts streams from various
                services and pipes them into a video player of choice.
            """),
            output
        )
        self.assertIn(
            dedent("""
                For more in-depth documentation see:
                  https://streamlink.github.io

                Please report broken plugins or bugs to the issue tracker on Github:
                  https://github.com/streamlink/streamlink/issues
            """),
            output
        )

    @patch("sys.stdout")
    @patch("sys.argv", ["streamlink", "--plugins"])
    def test_print_plugins(self, mock_stdout):
        self.subject()
        self.assertEqual(self.get_stdout(mock_stdout), "Loaded plugins: testplugin\n")

    @patch("sys.stdout")
    @patch("sys.argv", ["streamlink", "--plugins", "--json"])
    def test_print_plugins_json(self, mock_stdout):
        self.subject()
        self.assertEqual(self.get_stdout(mock_stdout), """[\n  "testplugin"\n]\n""")



from streamlink.plugins.clubbingtv import ClubbingTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlClubbingTV(PluginCanHandleUrl):
    __plugin__ = ClubbingTV

    should_match = [
        "https://www.clubbingtv.com/live",
        "https://www.clubbingtv.com/video/play/3950/moonlight/",
        "https://www.clubbingtv.com/video/play/2897",
        "https://www.clubbingtv.com/tomer-s-pick/",
    ]



from streamlink.plugins.facebook import Facebook
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlFacebook(PluginCanHandleUrl):
    __plugin__ = Facebook

    should_match = [
        "https://www.facebook.com/nos/videos/1725546430794241/",
        "https://www.facebook.com/nytfood/videos/1485091228202006/",
        "https://www.facebook.com/SporTurkTR/videos/798553173631138/",
        "https://www.facebook.com/119555411802156/posts/500665313691162/",
        "https://www.facebookwkhpilnemxj7asaniu7vnjjbiltxjqhye3mhbshg7kx5tfyd.onion/SporTurkTR/videos/798553173631138/",
    ]

    should_not_match = [
        "https://www.facebook.com",
    ]



"""
$description German news and documentaries TV channel, owned by Axel Springer SE.
$url welt.de
$type live, vod
$region Germany
"""

import re
from urllib.parse import quote

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r"https?://(\w+\.)?welt\.de/?"
))
class Welt(Plugin):
    _url_vod = "https://www.welt.de/onward/video/play/{0}"
    _schema = validate.Schema(
        validate.parse_html(),
        validate.xml_findtext(".//script[@type='application/json'][@data-content='VideoPlayer.Config']"),
        validate.parse_json(),
        validate.get("sources"),
        validate.filter(lambda obj: obj["extension"] == "m3u8"),
        validate.get((0, "src"))
    )

    def _get_streams(self):
        hls_url = self.session.http.get(self.url, schema=self._schema)

        if "mediathek" in self.url.lower():
            url = self._url_vod.format(quote(hls_url, safe=""))
            hls_url = self.session.http.get(url, headers={"Referer": self.url}).url

        return HLSStream.parse_variant_playlist(self.session, hls_url, headers={"Referer": self.url})


__plugin__ = Welt



import argparse
import re

_filesize_re = re.compile(r'''
    (?P<size>\d+(\.\d+)?)
    (?P<modifier>[Kk]|[Mm])?
    (?:[Bb])?
''', re.VERBOSE)
_keyvalue_re = re.compile(r'(?P<key>[^=]+)\s*=\s*(?P<value>.*)')


def boolean(value):
    truths = ['yes', '1', 'true', 'on']
    falses = ['no', '0', 'false', 'off']
    if value.lower() not in truths + falses:
        raise argparse.ArgumentTypeError('{0} was not one of {{{1}}}'.format(
            value, ', '.join(truths + falses)))

    return value.lower() in truths


def comma_list(values):
    return [val.strip() for val in values.split(',')]


def comma_list_filter(acceptable):
    def func(p):
        values = comma_list(p)
        return list(filter(lambda v: v in acceptable, values))

    return func


def filesize(value):
    match = _filesize_re.match(value)
    if not match:
        raise ValueError

    size = float(match.group('size'))
    if not size:
        raise ValueError

    modifier = match.group('modifier')
    if modifier in ('M', 'm'):
        size *= 1024 * 1024
    elif modifier in ('K', 'k'):
        size *= 1024

    return num(int, min=0)(size)


def keyvalue(value):
    match = _keyvalue_re.match(value)
    if not match:
        raise ValueError

    return match.group('key', 'value')


def num(type, min=None, max=None):
    def func(value):
        value = type(value)

        if min is not None and not (value > min):
            raise argparse.ArgumentTypeError(
                '{0} value must be more than {1} but is {2}'.format(
                    type.__name__, min, value
                )
            )

        if max is not None and not (value <= max):
            raise argparse.ArgumentTypeError(
                '{0} value must be at most {1} but is {2}'.format(
                    type.__name__, max, value
                )
            )

        return value

    func.__name__ = type.__name__

    return func


__all__ = [
    'boolean', 'comma_list', 'comma_list_filter', 'filesize', 'keyvalue',
    'num'
]



from streamlink.plugins.swisstxt import Swisstxt
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSwisstxt(PluginCanHandleUrl):
    __plugin__ = Swisstxt

    should_match = [
        "http://www.srf.ch/sport/resultcenter/tennis?eventId=338052",
        "http://live.rsi.ch/tennis.html?eventId=338052",
        "http://live.rsi.ch/sport.html?eventId=12345"
    ]

    should_not_match = [
        # regular srgssr sites
        "http://srf.ch/play/tv/live",
        "http://www.rsi.ch/play/tv/live#?tvLiveId=livestream_La1",
        "http://rsi.ch/play/tv/live?tvLiveId=livestream_La1",
        "http://www.rtr.ch/play/tv/live",
        "http://rtr.ch/play/tv/live",
        "http://rts.ch/play/tv/direct#?tvLiveId=3608506",
        "http://www.srf.ch/play/tv/live#?tvLiveId=c49c1d64-9f60-0001-1c36-43c288c01a10",
        "http://www.rts.ch/sport/direct/8328501-tennis-open-daustralie.html",
        "http://www.rts.ch/play/tv/tennis/video/tennis-open-daustralie?id=8328501",
    ]



from streamlink.plugins.telefe import Telefe
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTelefe(PluginCanHandleUrl):
    __plugin__ = Telefe

    should_match = [
        "https://mitelefe.com/vivo",
        "https://mitelefe.com/vivo/",
    ]

    should_not_match = [
        "http://telefe.com/pone-a-francella/temporada-1/programa-01/",
        "http://telefe.com/los-simuladores/temporada-1/capitulo-01/",
        "http://telefe.com/dulce-amor/capitulos/capitulo-01/",
        "http://telefe.com/",
        "http://www.telefeinternacional.com.ar/",
        "http://marketing.telefe.com/",
    ]



from streamlink.plugins.raiplay import RaiPlay
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRaiPlay(PluginCanHandleUrl):
    __plugin__ = RaiPlay

    should_match = [
        "http://www.raiplay.it/dirette/rai1",
        "http://www.raiplay.it/dirette/rai2",
        "http://www.raiplay.it/dirette/rai3",
        "http://raiplay.it/dirette/rai3",
        "https://raiplay.it/dirette/rai3",
        "http://www.raiplay.it/dirette/rainews24",
        "https://www.raiplay.it/dirette/rainews24",
    ]



from streamlink.plugins.nhkworld import NHKWorld
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNHKWorld(PluginCanHandleUrl):
    __plugin__ = NHKWorld

    should_match = [
        'https://www3.nhk.or.jp/nhkworld/en/live/',
    ]



from streamlink.plugins.zdf_mediathek import ZDFMediathek
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlZDFMediathek(PluginCanHandleUrl):
    __plugin__ = ZDFMediathek

    should_match = [
        'http://www.zdf.de/live-tv',
        'https://www.zdf.de/sender/zdf/zdf-live-beitrag-100.html',
        'https://www.zdf.de/sender/zdfneo/zdfneo-live-beitrag-100.html',
        'https://www.zdf.de/sender/3sat/3sat-live-beitrag-100.html',
        'https://www.zdf.de/sender/phoenix/phoenix-live-beitrag-100.html',
        'https://www.zdf.de/sender/arte/arte-livestream-100.html',
        'https://www.zdf.de/sender/kika/kika-live-beitrag-100.html',
        'https://www.zdf.de/dokumentation/zdfinfo-doku/zdfinfo-live-beitrag-100.html',
        'https://www.zdf.de/comedy/heute-show/videos/diy-hazel-habeck-100.html',
        'https://www.zdf.de/nachrichten/heute-sendungen/so-wird-das-wetter-102.html',
        'https://www.3sat.de/wissen/nano',
    ]



import unittest

from streamlink.plugins.albavision import Albavision
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAlbavision(PluginCanHandleUrl):
    __plugin__ = Albavision

    should_match = [
        "http://antena7.com.do/envivo-canal-7/",
        "http://www.antena7.com.do/envivo-canal-7/",
        "https://antena7.com.do/envivo-canal-7/",
        "https://www.antena7.com.do/envivo-canal7",
        "https://www.antena7.com.do/envivo-canal7/",
        "https://www.antena7.com.do/envivo-canal7#",
        "https://www.antena7.com.do/envivo-canal-7#",
        "https://www.antena7.com.do/en-vivo-canal-99/",
        "https://www.antena7.com.do/en-vivo-canal-99#",
        # All channel URLs from supported sites
        "https://www.antena7.com.do/envivo-canal-7/",
        "https://www.antena7.com.do/envivo-canal-21/",
        "https://www.atv.pe/envivo-atv",
        "https://www.atv.pe/envivo-atvmas",
        "https://www.c9n.com.py/envivo/",
        "https://www.canal10.com.ni/envivo/",
        "https://www.canal12.com.sv/envivo/",
        "https://www.chapintv.com/envivo-canal-3/",
        "https://www.chapintv.com/envivo-canal-7/",
        "https://www.chapintv.com/envivo-canal-23/",
        "https://www.elnueve.com.ar/en-vivo/",
        "https://www.redbolivision.tv.bo/envivo-canal-5/",
        "https://www.redbolivision.tv.bo/upptv/",
        "https://www.repretel.com/envivo-canal2/",
        "https://www.repretel.com/envivo-canal4/",
        "https://www.repretel.com/envivo-canal6/",
        "https://www.repretel.com/en-vivo-canal-11/",
        "https://www.rts.com.ec/envivo/",
        "https://www.snt.com.py/envivo/",
        "https://www.tvc.com.ec/envivo/",
        "https://www.vtv.com.hn/envivo/",
    ]
    should_not_match = [
        "https://fake.antena7.com.do/envivo-canal-7/",
        "https://www.antena7.com.do/envivo-canal123",
        "https://www.antena7.com.do/envivo-canal123/",
        "https://www.antena7.com.do/envivo-canal-123",
        "https://www.antena7.com.do/envivo-canal-123/",
        "https://www.antena7.com.do/envivo-canal-123#",
        "https://www.antena7.com.do/envivo-canalabc",
        "https://www.antena7.com.do/envivo-canal-abc",
        "https://www.antena7.com.do/envivo-canal-7/extra",
        "https://www.antena7.com.do/envivo-canal-7#extra",
        "https://www.antena7.com.do/something",
    ]


class TestPluginAlbavision(unittest.TestCase):
    def test_transform(self):
        token = Albavision.transform_token("6b425761cc8a86569b1a05a9bf1870c95fca717dOK", 436171)
        assert token == "6b425761cc8a86569b1a05a9bf1870c95fca717d"
